from . import aws_backend
from . import aws_util
from . import util
from . import local_backend
from . import backend  # TODO: remove?

from .ncluster import get_backend
from .ncluster import set_backend
from .ncluster import running_locally

from .ncluster import use_aws
from .ncluster import use_local

from .ncluster import make_task
from .ncluster import make_job
from .ncluster import make_run
from .ncluster import get_zone
from .ncluster import get_region
from .ncluster import set_logdir_root
from .ncluster import get_logdir_root
from .ncluster import add_authorized_keys  # is this still used?
from .aws_util import running_on_aws

from ._version import __version__

print(f"ncluster version {__version__}")
# set default backend from environment
set_backend('aws')


if not util.is_set('NCLUSTER_DISABLE_PDB_HANDLER'):
  util.install_pdb_handler()  # CTRL+\ drops into pdb



"""Local implementation of backend.py using separate tmux sessions for jobs.

Not thread-safe.
"""

import glob
import os
import shlex
import socket
import time
from typing import List

from ncluster import ncluster_globals
from . import backend
from . import util

TASKDIR_ROOT = '/tmp/ncluster/task'
SCRATCH_ROOT = '/tmp/ncluster/scratch'
#LOGDIR_ROOT = os.environ['HOME'] + '/ncluster/runs'  # use ~ instead of /tmp because /tmp gets wiped
DEFAULT_LOGDIR_ROOT = '/ncluster/runs'


# todo: tmux session names are backwards from AWS job names (runname-jobname)
# TODO: add kwargs so that tmux backend can be drop-in replacement


# TODO: rename extra_kwargs to kwargs everywhere
class Task(backend.Task):
  """Local tasks interact with tmux session where session name is derived
  from job name, and window names are task ids."""
  tmux_window_id: int
  tmux_available_window_ids: List[int]

  def __init__(self, name, *, tmux_session, install_script='', job=None,
               **kwargs):

    self.homedir = os.environ['HOME']
    self._cmd_fn = None
    self._cmd = None
    self._status_fn = None  # location of output of last status
    self._out_fn = None

    self._can_run = False
    self.tmux_session = tmux_session
    self.tmux_window_id = 0
    self.tmux_available_window_ids = [0]

    self.name = name
    self.install_script = install_script
    self.job = job
    self.kwargs = kwargs

    # local servers sometimes listen only on localhost (TensorBoard), and sometimes only on
    # externally assigned ip address from gethostbyname (Ray), must choose one, so use the localhost for TB compatibility
    # https://github.com/ray-project/ray/issues/1677
    self.public_ip = socket.gethostbyname(socket.gethostname())
    #  self.public_ip = '127.0.0.1'
    self.ip = self.public_ip

    self.connect_instructions = 'tmux a -t ' + self.tmux_session

    # task current dir
    print('name is', name)
    # tmpdir = f"{util.reverse_taskname(name)}.{os.getpid()}.{util.now_micros()}"
    launch_id = util.random_id()
    self.taskdir = f"{TASKDIR_ROOT}/{name}-{launch_id}"
    self.local_scratch = f"{SCRATCH_ROOT}/{name}-{launch_id}"
    self.remote_scratch = f"{SCRATCH_ROOT}/{name}-{launch_id}"

    self.log(f"Creating taskdir {self.taskdir}")
    self._run_raw('mkdir -p ' + self.taskdir)

    self.log(f"Creating scratch {self.local_scratch}")
    self._run_raw('rm -Rf ' + self.local_scratch)
    self._run_raw('mkdir -p ' + self.local_scratch)
    self._run_raw('mkdir -p ' + self.remote_scratch)
    self.run_counter = 0

    self._cwd = self.taskdir
    self._can_run = True
    self.run('cd ' + self.taskdir)

    print("Running install script " + install_script)
    self.install_script = install_script
    for line in install_script.split('\n'):
      self.run(line)

  def run(self, cmd, non_blocking=False, ignore_errors=False, **_kwargs):

    if util.is_set('NCLUSTER_RUN_WITH_OUTPUT_ON_FAILURE'):
      # HACK
      if not util.is_bash_builtin(cmd) or True:
        return self._run_with_output_on_failure(cmd, non_blocking, ignore_errors, **_kwargs)
      else:
        self.log("Found bash built-in, using regular run")

    if not self._can_run:
      assert False, "Using .run before initialization finished"
    if '\n' in cmd:
      cmds = cmd.split('\n')
      self.log(
        f"Running {len(cmds)} commands at once, returning status of last")
      status = -1
      for subcmd in cmds:
        status = self.run(subcmd)
      return status

    cmd = cmd.strip()
    if not cmd or cmd.startswith('#'):  # ignore empty/commented out lines
      return -1
    self.run_counter += 1
    self.log("tmux> %s", cmd)

    self._cmd = cmd
    self._cmd_fn = f'{self.local_scratch}/{self.run_counter}.cmd'
    self._status_fn = f'{self.remote_scratch}/{self.run_counter}.status'
    assert not os.path.exists(self._status_fn)

    cmd = util.shell_strip_comment(cmd)
    # assert '&' not in cmd, f"cmd {cmd} contains &, that breaks things"

    self.write(self._cmd_fn, cmd + '\n')
    modified_cmd = f'{cmd} ; echo $? > {self._status_fn}'
    modified_cmd = shlex.quote(modified_cmd)

    tmux_window = self.tmux_session+':'+str(self.tmux_window_id)
    tmux_cmd = f'tmux send-keys -t {tmux_window} {modified_cmd} Enter'
    self._run_raw(tmux_cmd, ignore_errors=ignore_errors)
    if non_blocking:
      return 0

    if not self.wait_for_file(self._status_fn, max_wait_sec=60):
      self.log(f"Retrying waiting for {self._status_fn}")
    while not self.exists(self._status_fn):
      self.log(f"Still waiting for {cmd}")
      self.wait_for_file(self._status_fn, max_wait_sec=60)
    contents = self.read(self._status_fn)

    # if empty wait a bit to allow for race condition
    if len(contents) == 0:
      time.sleep(0.01)
    status = int(open(self._status_fn).read().strip())
    self.last_status = status

    if status != 0:
      if not ignore_errors:
        raise RuntimeError(f"Command {cmd} returned status {status}")
      else:
        self.log(f"Warning: command {cmd} returned status {status}")

    return status

  def join(self, ignore_errors=False):
    """Waits until last executed command completed."""
    assert self._status_fn, "Asked to join a task which hasn't had any commands executed on it"
    check_interval = 0.2
    status_fn = self._status_fn
    if not self.wait_for_file(status_fn, max_wait_sec=30):
      self.log(f"Retrying waiting for {status_fn}")
    while not self.exists(status_fn):
      self.log(f"Still waiting for {self._cmd}")
      self.wait_for_file(status_fn, max_wait_sec=30)
    contents = self.read(status_fn)

    # if empty wait a bit to allow for race condition
    if len(contents) == 0:
      time.sleep(check_interval)
      contents = self.read(status_fn)
    status = int(contents.strip())
    self.last_status = status

    if status != 0:
      extra_msg = '(ignoring error)' if ignore_errors else '(failing)'
      if util.is_set('NCLUSTER_RUN_WITH_OUTPUT_ON_FAILURE'):
        self.log(
          f"Start failing output {extra_msg}: \n{'*'*80}\n\n '{self.read(self._out_fn)}'")
        self.log(f"\n{'*'*80}\nEnd failing output")
      if not ignore_errors:
        raise RuntimeError(f"Command {self._cmd} returned status {status}")
      else:
        self.log(f"Warning: command {self._cmd} returned status {status}")

    return status

  def switch_window(self, window_id: int):
    """
    Switches currently active tmux window for given task. 0 is the default window
    Args:
      window_id: integer id of tmux window to use
    """

    # windows are numbered sequentially 0, 1, 2, ...
    # create any missing windows and make them point to the same directory
    if window_id not in self.tmux_available_window_ids:
      for i in range(max(self.tmux_available_window_ids)+1, window_id+1):
        self._run_raw(f'tmux new-window -t {self.tmux_session} -d')

        tmux_window = self.tmux_session + ':' + str(i)
        cmd = shlex.quote(f'cd {self.taskdir}')
        tmux_cmd = f'tmux send-keys -t {tmux_window} {cmd} Enter'
        self._run_raw(tmux_cmd)
        self.tmux_available_window_ids.append(i)

    self.tmux_window_id = window_id

  # This is a future "run" command, will become "run" once all cases are checked
  def _run_with_output_on_failure(self, cmd, non_blocking=False, ignore_errors=False, **_kwargs) -> str:
    if not self._can_run:
      assert False, "Using .run before initialization finished"
    if '\n' in cmd:
      cmds = cmd.split('\n')
      self.log(
        f"Running {len(cmds)} commands at once, returning status of last")
      status = -1
      for subcmd in cmds:
        status = self.run(subcmd)
      return status

    cmd = cmd.strip()
    if not cmd or cmd.startswith('#'):  # ignore empty/commented out lines
      return ''
    self.run_counter += 1
    self.log("tmux> %s", cmd)

    self._cmd = cmd
    self._cmd_fn = f'{self.local_scratch}/{self.run_counter}.cmd'
    self._status_fn = f'{self.remote_scratch}/{self.run_counter}.status'
    self._out_fn = f'{self.remote_scratch}/{self.run_counter}.out'
    assert not os.path.exists(self._status_fn)

    cmd = util.shell_strip_comment(cmd)
    # assert '&' not in cmd, f"cmd {cmd} contains &, that breaks things"

    self.write(self._cmd_fn, cmd + '\n')
    #  modified_cmd = f'{cmd} ; echo $? > {self._status_fn}'
    modified_cmd = f'{cmd} > >(tee -a {self._out_fn}) 2> >(tee -a {self._out_fn} >&2); echo $? > {self._status_fn}'
    modified_cmd = shlex.quote(modified_cmd)

    tmux_window = self.tmux_session+':'+str(self.tmux_window_id)
    tmux_cmd = f'tmux send-keys -t {tmux_window} {modified_cmd} Enter'
    self._run_raw(tmux_cmd)
    if non_blocking:
      return ''

    if not self.wait_for_file(self._status_fn, max_wait_sec=60):
      self.log(f"Retrying waiting for {self._status_fn}")
    while not self.exists(self._status_fn):
      self.log(f"Still waiting for {cmd}")
      self.wait_for_file(self._status_fn, max_wait_sec=60)
    contents = self.read(self._status_fn)

    # if empty wait a bit to allow for race condition
    if len(contents) == 0:
      time.sleep(0.01)
    status = int(open(self._status_fn).read().strip())
    self.last_status = status

    if status != 0:
      extra_msg = '(ignoring error)' if ignore_errors else '(failing)'
      self.log(
        f"Start failing output {extra_msg}: \n{'*'*80}\n\n '{self.read(self._out_fn)}'")
      self.log(f"\n{'*'*80}\nEnd failing output")
      if not ignore_errors:
        raise RuntimeError(f"Command {cmd} returned status {status}")
      else:
        self.log(f"Warning: command {cmd} returned status {status}")

    return self.read(self._out_fn)

  def _run_raw(self, cmd, ignore_errors=False):
    """Runs command directly, skipping tmux interface"""
    # TODO: capture stdout/stderr for feature parity with aws_backend
    result = os.system(cmd)
    if result != 0:
      if ignore_errors:
        self.log(f"command ({cmd}) failed.")
        assert False, "_run_raw failed"

  def rsync(self, local_fn, remote_fn=None):
    self.upload(local_fn, remote_fn)
  
  def upload(self, local_fn, remote_fn=None, dont_overwrite=False):
    """Uploads file to remote instance. If location not specified, dumps it
    into default directory. Creates missing directories in path name."""

    # support wildcard through glob
    if '*' in local_fn:
      for local_subfn in glob.glob(local_fn):
        self.upload(local_subfn)
      return

    if remote_fn is None:
      remote_fn = os.path.basename(local_fn)

    if dont_overwrite and self.exists(remote_fn):
      self.log("Remote file %s exists, skipping" % (remote_fn,))
      return

    if not remote_fn.startswith('/'):
      remote_fn = self.taskdir + '/' + remote_fn

    remote_fn = remote_fn.replace('~', self.homedir)
    self.log('uploading ' + local_fn + ' to ' + remote_fn)

    local_fn = os.path.abspath(local_fn)
    self._run_raw("cp -R %s %s" % (local_fn, remote_fn))

  def download(self, remote_fn, local_fn='.'):
    if local_fn == '.':
      local_fn = self._cwd
    #    self.log("downloading %s to %s" % (remote_fn, local_fn))
    if not remote_fn.startswith('/'):
      remote_fn = self._cwd + '/' + remote_fn
    if self.exists(remote_fn):
      os.system(f'cp {remote_fn} {local_fn}')
    else:
      raise RuntimeError(f"No such file {remote_fn}")

  def exists(self, remote_fn):
    return os.path.exists(remote_fn)

  def read(self, remote_fn):
    tmp_fn = self.local_scratch + '/' + str(util.now_micros())
    self.download(remote_fn, tmp_fn)
    return open(tmp_fn).read()

  def write(self, remote_fn, contents):
    def make_temp_fn():
      """Returns temporary filename for this task."""
      return self.local_scratch + '/write.' + str(util.now_micros())

    tmp_fn = make_temp_fn()
    open(tmp_fn, 'w').write(contents)
    self.upload(tmp_fn, remote_fn)

  # don't include file streaming for now
  # the issue is that file streaming by default turns on 4K buffering, which makes
  # streaming a lot less useful. Similar buffering is turned on for piping commands
  # https://unix.stackexchange.com/questions/25372/turn-off-buffering-in-pipe
  # def file_stream(self, fn: str) -> None:
  #   #    if not fn.startswith('/'):
  #   #      fn = self.taskdir + '/' + fn
  #
  #   if not os.path.exists(fn):
  #     os.system('mkdir -p ' + os.path.dirname(os.path.abspath(fn)))
  #     os.system('touch ' + fn)
  #
  #   p = subprocess.Popen(['tail', '-f', fn], stdout=subprocess.PIPE)
  #
  #   for line in iter(p.stdout.readline, ''):
  #     sys.stdout.write(line.decode('ascii', errors='ignore'))

  @property
  def logdir(self):
    """Returns logging directory, creating one if necessary. See "Logdir" section  of design doc on naming convention."""

    run_name = ncluster_globals.get_run_for_task(self)
    logdir = ncluster_globals.get_logdir(run_name)
    if logdir:
      return logdir

    # create logdir. Only single task in a group creates the logdir
    if ncluster_globals.is_chief(self, run_name):
      chief = self
    else:
      chief = ncluster_globals.get_chief(run_name)

    chief.setup_logdir()
    return ncluster_globals.get_logdir(run_name)
   # release lock

  def setup_logdir(self):
    # todo: locking on logdir creation

    """Create logdir for task/job/run. No-op if the task is not chief (0'th task of 0'th job of run)
    """
    run_name = ncluster_globals.get_run_for_task(self)
    self.log("Creating logdir for run "+run_name)
    logdir_root = ncluster_globals.LOGDIR_ROOT
    assert logdir_root, "LOGDIR_ROOT not set, make sure you have called ncluster.set_backend()"

    self.run(f'mkdir -p {logdir_root}')
    find_command = f'find {logdir_root} -maxdepth 1 -type d'

    stdout, stderr = self.run_with_output(find_command)
    logdir = f"{logdir_root}/{run_name}"

    counter = 0
    while logdir in stdout:
      counter += 1
      new_logdir = f'{logdir_root}/{run_name}.{counter:02d}'
      self.log(f'Warning, logdir {logdir} exists, deduping to {new_logdir}')
      logdir = new_logdir
    self.run(f'mkdir -p {logdir}')

    ncluster_globals.set_logdir(run_name, logdir)
    return logdir


class Job(backend.Job):
  pass


class Run:
  """Run is a collection of jobs that share state. IE, training run will contain gradient worker job, parameter
  server job, and TensorBoard visualizer job. These jobs will use the same shared directory to store checkpoints and
  event files.
  :ivar aws_placement_group_name: somedoc
  """
  jobs: List[Job]

  def __init__(self, name='', **kwargs):
    """Creates a run. If install_script is specified, it's used as default
    install_script for all jobs (can be overridden by Job constructor)"""

    self.name = name
    self.kwargs = kwargs

  @property
  def logdir(self):
    chief_task = ncluster_globals.get_chief(self.name)
    return chief_task.logdir

  # TODO: currently this is synchronous, use non_blocking wrapper like in Job to parallelize methods
  def run(self, *args, **kwargs):
    """Runs command on every job in the run."""

    for job in self.jobs:
      job.run(*args, **kwargs)

  def run_with_output(self, *args, **kwargs):
    """Runs command on every first job in the run, returns stdout."""
    for job in self.jobs:
      job.run_with_output(*args, **kwargs)

  def _run_raw(self, *args, **kwargs):
    """_run_raw on every job in the run."""
    for job in self.jobs:
      job._run_raw(*args, **kwargs)

  def upload(self, *args, **kwargs):
    """Runs command on every job in the run."""
    for job in self.jobs:
      job.upload(*args, **kwargs)

  def rsync(self, *args, **kwargs):
    """Runs command on every job in the run."""
    for job in self.jobs:
      job.rsync(*args, **kwargs)

  def make_job(self, name='', **kwargs):
    return make_job(name+'.'+self.name, run_name=self.name, **kwargs)


def make_task(name='',
              run_name='',
              **kwargs) -> Task:
  """Create task, also create dummy run if not specified."""
  ncluster_globals.task_launched = True

  name = ncluster_globals.auto_assign_task_name_if_needed(name)

  # tmux can't use . for session names
  tmux_session = name.replace('.', '=')
  tmux_window_id = 0
  util.log(f'killing session {tmux_session}')

  if not util.is_set("NCLUSTER_NOKILL_TMUX"):
    os.system(f'tmux kill-session -t {tmux_session}')
  os.system(f'tmux new-session -s {tmux_session} -n {tmux_window_id} -d')

  task = Task(name,
              tmux_session=tmux_session,  # propagate optional args
              run_name=run_name,
              **kwargs)
  ncluster_globals.register_task(task, run_name)
  return task


def make_job(name="",
             num_tasks=1,
             run_name="",
             install_script='',
             **kwargs
             ) -> backend.Job:
  assert num_tasks > 0, f"Can't create job with {num_tasks} tasks"

  name = ncluster_globals.auto_assign_job_name_if_needed(name)
  util.validate_ncluster_job_name(name)
  tasks = [make_task(f"{i}.{name}",
                     run_name=run_name,
                     install_script=install_script,
                     **kwargs
                     ) for i in range(num_tasks)]

  job = backend.Job(name=name, tasks=tasks, **kwargs)
  return job


def make_run(name) -> Run:
  run = Run(name)
  return run



"""Methods used in aws_backend, but also useful for standalone prototyping in Jupyter."""

import os
import re
import subprocess
import sys
import time
from collections import OrderedDict

import botocore.exceptions
import paramiko
from operator import itemgetter

from typing import Iterable, List, Dict, Optional

from boto3_type_annotations.ec2 import SecurityGroup, Vpc, Subnet, InternetGateway, PlacementGroup, Image, \
  Instance, KeyPairInfo
from boto3.resources.collection import ResourceCollection

from boto3_type_annotations.ec2 import ServiceResource as EC2_ServiceResource
from boto3_type_annotations.ec2 import Client as EC2_Client
from boto3_type_annotations.sts import Client as STS_Client

import boto3

from . import util

from .util import VALID_REGIONS

EMPTY_NAME = "noname"  # name to use when name attribute is missing on AWS
RETRY_INTERVAL_SEC = 1  # how long to wait before retries
RETRY_TIMEOUT_SEC = 60  # how long to wait before retrying fails
DEFAULT_PREFIX = 'ncluster'
PRIVATE_KEY_LOCATION = os.environ['HOME'] + '/.ncluster'
DUPLICATE_CHECKING = False

# allow referring to local methods as u.<method>, for easier moving of code between this module and others
u = sys.modules[__name__]


# Can't annotate boto3 return types because they are missing stubs
# https://github.com/boto/boto3/issues/1055
# https://stackoverflow.com/questions/52087307/adding-type-hinting-to-functions-that-return-boto3-objects

def get_vpc() -> Vpc:
  """
  Returns current VPC (ec2.Vpc object)
  https://boto3.readthedocs.io/en/latest/reference/services/ec2.html#vpc
  """

  return get_vpc_dict()[get_prefix()]


def get_security_group() -> SecurityGroup:
  """
  Returns current security group, ec2.SecurityGroup object
  https://boto3.readthedocs.io/en/latest/reference/services/ec2.html#securitygroup
"""
  return get_security_group_dict()[get_security_group_name()]


def get_security_group_nd() -> SecurityGroup:
  """Gets security group associated with non-default VPC"""
  return get_security_group_dict()[get_security_group_nd_name()]


def get_subnet() -> Subnet:
  zone = get_zone()
  assert zone, f"Subnet is only defined for zone, current zone is '{zone}', use $NCLUSTER_ZONE to define"
  subnet_dict = get_subnet_dict()
  assert zone in subnet_dict, f"zone '{zone}' is not in subnet_dict, available zones are {subnet_dict.keys()}"
  return subnet_dict[zone]


def get_vpc_dict() -> Dict[str, Vpc]:
  """Returns dictionary of named VPCs {name: vpc}

  Assert fails if there's more than one VPC with same name."""

  client = get_ec2_client()
  response = client.describe_vpcs()
  assert is_good_response(response)

  result = OrderedDict()
  ec2 = get_ec2_resource()
  for vpc_response in response['Vpcs']:
    key = get_name(vpc_response.get('Tags', []))
    if not key or key == EMPTY_NAME:  # skip VPC's that don't have a name assigned
      continue

    if key in result:
      util.log(f"Warning: Duplicate VPC group {key} in {response}")
      if DUPLICATE_CHECKING:
        assert False
    result[key] = ec2.Vpc(vpc_response['VpcId'])

  return result


def get_default_vpc() -> Vpc:
  """
  Return default VPC or none if not present

  """
  ec2 = get_ec2_resource()
  for vpc in ec2.vpcs.all():
    if vpc.is_default:
      return vpc


def get_subnet_dict() -> Dict[str, Subnet]:
  """Returns dictionary of "availability zone" -> subnet for current VPC."""
  subnet_dict = {}
  vpc = get_default_vpc()
  for subnet in vpc.subnets.all():
    zone = subnet.availability_zone
    assert zone not in subnet_dict, "More than one subnet in %s, why?" % (zone,)
    subnet_dict[zone] = subnet
  return subnet_dict


def get_gateway_dict(vpc) -> Dict[str, InternetGateway]:
  """Returns dictionary of named gateways for given VPC {name: gateway}"""
  return {get_name(gateway): gateway for
          gateway in vpc.internet_gateways.all()}


def get_efs_dict() -> Dict[str, str]:
  """Returns dictionary of {efs_name: efs_id}"""
  # there's no EC2 resource for EFS objects, so return EFS_ID instead
  # https://stackoverflow.com/questions/47870342/no-ec2-resource-for-efs-objects

  efs_client = get_efs_client()
  response = call_with_retries(efs_client.describe_file_systems,
                               'efs_client.describe_file_systems')
  assert is_good_response(response)
  result = OrderedDict()
  for efs_response in response['FileSystems']:
    fs_id = efs_response['FileSystemId']

    tag_response = call_with_retries(efs_client.describe_tags,
                                     "efs_client.describe_tags",
                                     FileSystemId=fs_id, retry_interval_sec=2)
    assert is_good_response(tag_response)
    key = get_name(tag_response['Tags'])
    if not key or key == EMPTY_NAME:  # skip EFS's without a name
      continue
    assert key not in result
    result[key] = fs_id

  return result


def get_placement_group_dict() -> Dict[str, PlacementGroup]:
  """Returns dictionary of {placement_group_name: (state, strategy)}"""

  client = get_ec2_client()
  response = client.describe_placement_groups()
  assert is_good_response(response)

  result = OrderedDict()
  ec2 = get_ec2_resource()
  for placement_group_response in response['PlacementGroups']:
    key = placement_group_response['GroupName']
    if key in result:
      util.log(f"Warning: Duplicate placement_group group {key}")
      if DUPLICATE_CHECKING:
        assert False
    result[key] = ec2.PlacementGroup(key)
  return result


def get_security_group_dict() -> Dict[str, SecurityGroup]:
  """Returns dictionary of named security groups {name: securitygroup}."""

  client = get_ec2_client()
  response = client.describe_security_groups()
  assert is_good_response(response)

  result = OrderedDict()
  ec2 = get_ec2_resource()
  for security_group_response in response['SecurityGroups']:
    key = get_name(security_group_response.get('Tags', []))
    if not key or key == EMPTY_NAME:
      continue  # ignore unnamed security groups
    #    key = security_group_response['GroupName']
    if key in result:
      util.log(f"Warning: Duplicate security group {key}")
      if DUPLICATE_CHECKING:
        assert key not in result, ("Duplicate security group " + key)
    result[key] = ec2.SecurityGroup(security_group_response['GroupId'])

  return result


def get_keypair_dict() -> Dict[str, KeyPairInfo]:
  """Returns dictionary of {keypairname: keypair}"""

  client = get_ec2_client()
  response = client.describe_key_pairs()
  assert is_good_response(response)

  result = {}
  ec2 = get_ec2_resource()
  for keypair in response['KeyPairs']:
    keypair_name = keypair.get('KeyName', '')
    if keypair_name in result:
      util.log(f"Warning: Duplicate key {keypair_name}")
    if DUPLICATE_CHECKING:
      assert keypair_name not in result, "Duplicate key " + keypair_name
    result[keypair_name] = ec2.KeyPair(keypair_name)
  return result


def get_prefix() -> str:
  """Global prefix to identify ncluster created resources name used to identify ncluster created resources,
  (name of EFS, VPC, keypair prefixes), can be changed through $NCLUSTER_PREFIX for debugging purposes. """

  name = os.environ.get('NCLUSTER_PREFIX', DEFAULT_PREFIX)
  if name != DEFAULT_PREFIX:
    validate_prefix(name)
  return name


def get_account_number() -> str:
  start_time = time.time()
  success = False
  account_number = "Unknown"
  retry_times = 3
  for i in range(retry_times):
    if time.time() - start_time - RETRY_INTERVAL_SEC > RETRY_TIMEOUT_SEC:
      assert False, f"Timeout {RETRY_TIMEOUT_SEC} exceeded querying account number {retry_times} times"
    try:
      sts_client: STS_Client = boto3.client('sts')
      account_number = str(sts_client.get_caller_identity()['Account'])
      success = True
    except Exception as e:
      util.log(f'Exception in get_account_number {e}, retrying')
      if 'AWS_SECRET_ACCESS_KEY' not in os.environ:
        util.log(
          'AWS_SECRET_ACCESS_KEY not in env vars, configure your AWS credentials."')
      if 'AWS_ACCESS_KEY_ID' not in os.environ:
        util.log(
          'AWS_ACCESS_KEY_ID not in env vars, configure your AWS credentials."')
      time.sleep(RETRY_INTERVAL_SEC)
  public_key = os.environ.get('AWS_ACCESS_KEY_ID', None)
  if not success:
    assert False, f"Could access account, make sure you have correct credentials for region {get_region()} and key {public_key}"
  return account_number


def get_account_name() -> str:
  iam_client = boto3.client('iam')
  response = iam_client.list_account_aliases()
  aliases = response.get('AccountAliases', [])
  if aliases:
    return aliases[0]
  else:
    return ''


def get_iam_username() -> str:
  iam_client = boto3.client('iam')
  return iam_client.get_user()['User']['UserName']


def get_region() -> str:
  return get_session().region_name


def set_region(region: str) -> None:
  assert region in VALID_REGIONS, f"region '{region}' is not valid, valid regions are [{','.join(VALID_REGIONS)}]"
  util.set_env('AWS_DEFAULT_REGION', region)


def get_zone() -> str:
  """Returns current zone, or empty string if it's unset."""
  return util.get_env('NCLUSTER_ZONE')


def set_zone(zone: str) -> None:
  current_region = u.get_region()
  assert zone.startswith(current_region), f'Current zone "{zone}" ($NCLUSTER_ZONE) is not in current region "{current_region} ($AWS_DEFAULT_REGION), change either $NCLUSTER_ZONE or AWS_DEFAULT_REGION'
  util.set_env('NCLUSTER_ZONE', zone)


def get_zones() -> List[str]:
  client = get_ec2_client()
  response = client.describe_availability_zones()
  assert is_good_response(response)
  zones = []
  for avail_response in response['AvailabilityZones']:
    messages = avail_response['Messages']
    zone = avail_response['ZoneName']
    state = avail_response['State']
    assert not messages, f"zone {zone} is broken? Has messages {messages}"
    assert state == 'available', f"zone {zone} is broken? Has state {state}"
    zones.append(zone)
  return zones


def get_session():
  # in future can add aws profile support with Session(profile_name=...)
  return boto3.Session()


################################################################################
# keypairs
################################################################################
# For naming conventions, see
# https://docs.google.com/document/d/14-zpee6HMRYtEfQ_H_UN9V92bBQOt0pGuRKcEJsxLEA/edit#heading=h.45ok0839c0a

def get_keypair_name() -> str:
  """Returns current keypair name."""

  username = get_username()
  assert '-' not in username, "username must not contain -, change $USER"
  validate_aws_name(username)
  assert len(username) < 30  # to avoid exceeding AWS 127 char limit
  return get_prefix() + '-' + username


def get_keypair() -> KeyPairInfo:
  """Returns current keypair (ec2.KeyPairInfo)

  https://boto3.readthedocs.io/en/latest/reference/services/ec2.html#keypairinfo
  """

  return get_keypair_dict()[get_keypair_name()]


def get_keypair_fn() -> str:
  """Location of .pem file for current keypair"""

  keypair_name = get_keypair_name()
  account = get_account_number()
  region = get_region()
  fn = f'{PRIVATE_KEY_LOCATION}/{keypair_name}-{account}-{region}.pem'
  return fn


def get_vpc_name() -> str:
  return get_prefix()


def get_security_group_name() -> str:
  """Security group for default VPC"""
  return get_prefix()


def get_security_group_nd_name() -> str:
  """Security group for non-default VPC"""
  return get_prefix() + '_nd'


def get_gateway_name() -> str:
  return get_prefix()


def get_route_table_name() -> str:
  return get_prefix()


def get_efs_name() -> str:
  return get_prefix()


def get_username() -> str:
  assert 'USER' in os.environ, "why isn't USER defined?"
  return os.environ['USER']


def lookup_image(wildcard) -> Image:
  """Returns unique ec2.Image whose name matches wildcard
  lookup_ami('pytorch*').name => ami-29fa
  
  https://boto3.readthedocs.io/en/latest/reference/services/ec2.html#image

  Assert fails if multiple images match or no images match.
  """

  ec2 = get_ec2_resource()
  filter_ = {'Name': 'name', 'Values': [wildcard]}

  images = list(ec2.images.filter(Filters=[filter_]))

  # Note, can add filtering by Owners as follows
  #  images = list(ec2.images.filter_(Filters = [filter_], Owners=['self', 'amazon']))

  assert len(images) <= 1, "Multiple images match " + str(wildcard)
  assert len(images) > 0, f"No images match {str(wildcard)} in {get_region()}, check name and image permissions"
  return images[0]


def is_amazon_ami(image_name):
  """Heuristic to tell if given image name is Amazon Linux image."""
  image_name = image_name.lower()
  return 'amzn' in image_name or 'amazon' in image_name or image_name.startswith('dlami23-efa') or image_name.startswith('basic-efa')


def get_aws_username(instance) -> str:
  image_name = instance.image.name.lower()
  if util.get_env('NCLUSTER_SSH_USERNAME'):
    return util.get_env('NCLUSTER_SSH_USERNAME')

  if is_amazon_ami(image_name):
    print("Auto-detected Amazon Linux, using ec2-user ssh name")
    return 'ec2-user'
  else:
    return 'ubuntu'


def lookup_instance_exact(name: str, instance_type: str = '', image_name: str = '',
                          states: tuple = ('running', 'stopped', 'initializing'),
                          limit_to_current_user=False) -> Optional[Instance]:
  """Looks up AWS instance for given instance name using exact matching, like
   simple.worker. If no instance found in current AWS environment, returns None. """

  ec2 = get_ec2_resource()

  # https://github.com/boto/boto3/issues/2000
  # noinspection PyTypeChecker
  instances: ResourceCollection = ec2.instances.filter(
    Filters=[{'Name': 'instance-state-name', 'Values': states}])

  prefix = get_prefix()
  username = get_username()

  # look for an existing instance matching job, ignore instances launched
  # by different user or under different resource name
  result = []
  for i in instances.all():
    instance_name = get_name(i)
    if instance_name != name:
      continue

    if limit_to_current_user and i.key_name != get_keypair_name():
      continue

    seen_prefix, seen_username = parse_key_name(i.key_name)
    if prefix != seen_prefix:
      print(f"Found {name} launched under {seen_prefix}, ignoring because current ncluster prefix is {prefix}")
      continue
    if username != seen_username:
      print(f"Found {name} launched by {seen_username}, ignoring because current user is {username}")
      continue

    if instance_type:
      assert i.instance_type == instance_type, f"Found existing instance for job {name} but different instance type ({i.instance_type}) than requested ({instance_type}), terminate {name} first or use new task name."

    if image_name:
      assert i.image.name == image_name, f"Found existing instance for job {name} but launched with different image ({i.image.name}) than requested ({image_name}), terminate {name} first or use new task name."
    result.append(i)

    assert len(result) < 2, f"Found two instances with name {name}"
    if not result:
      return None
    else:
      return result[0]


def lookup_instance(fragment='', valid_states=('running',)):
  """Wrapper around lookup_instances that checks returns a single unambiguous matching instance."""

  instances = lookup_instances(fragment, valid_states=valid_states)
  assert instances, f"Didn't find any instances matching '{fragment}'"
  names = [get_name(i) for i in instances]
  assert len(instances) == 1, f"Found multiple instances matching fragment {fragment}: {','.join(names)}"
  return instances[0]


def lookup_instances(fragment='', *, verbose=True, filter_by_key=False, valid_states=('running',),
                     limit_to_current_user=False) -> List[Instance]:
  """Returns List of ec2.Instance object whose name contains fragment, in reverse order of launching (ie,
  most recent instance first). Optionally filters by key, only including instances launched with
  key_name matching current username.

  If fragment is wrapped in single quotes like 'somemachine', it strips quotes and will try to match somemachine exactly
  args:
    verbose: print information about all matching instances found

    filter_by_key  if True, ignore instances that are not launched with current
        user's default key
    limit_to_current_user Restrict result to instances that current user can ssh into

  """

  if fragment.startswith("'"):
    assert fragment.endswith("'")
    exact_match = True
    fragment = fragment[1:-1]
  else:
    exact_match = False

  def vprint(*args):
    if verbose:
      print(*args)

  region = get_region()
  client = get_ec2_client()
  ec2 = get_ec2_resource()
  response = client.describe_instances()
  assert is_good_response(response)

  instance_list = []
  for instance in ec2.instances.all():
    if instance.state['Name'] not in valid_states:
      continue
    if limit_to_current_user and instance.key_name != get_keypair_name():
      continue

    name = get_name(instance)
    if exact_match:
      if fragment == name:
        instance_list.append((util.toseconds(instance.launch_time), instance))
    else:
      if fragment in name:
        # the following can be re-enabled for broader match
        # or fragment in str(instance.public_ip_address) or
        # fragment in str(instance.id) or fragment in str(instance.private_ip_address)):
        instance_list.append((util.toseconds(instance.launch_time), instance))

  sorted_instance_list = reversed(sorted(instance_list, key=itemgetter(0)))
  filtered_instance_list = []  # filter by key
  vprint("Using region ", region)
  for (ts, instance) in sorted_instance_list:
    if filter_by_key and instance.key_name != get_keypair_name():
      vprint(f"Got key {instance.key_name}, expected {get_keypair_name()}")
      continue
    filtered_instance_list.append(instance)
  return filtered_instance_list


def ssh_to_task(task) -> paramiko.SSHClient:
  """Create ssh connection to task's machine

  returns Paramiko SSH client connected to host.

  """

  username = task.ssh_username
  hostname = task.public_ip
  ssh_key_fn = get_keypair_fn()
  print(f"ssh -i {ssh_key_fn} {username}@{hostname}")
  pkey = paramiko.RSAKey.from_private_key_file(ssh_key_fn)

  ssh_client = paramiko.SSHClient()
  ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
  assert ssh_client

  counter = 1
  while True:
    try:
      ssh_client.connect(hostname=hostname, username=username, pkey=pkey)
      if counter % 11 == 0:  # occasionally re-obtain public ip, machine could've gotten restarted
        hostname = task.public_ip
      break
    except Exception as e:
      print(
        f'{task.name}: Exception connecting to {hostname} via ssh (could be a timeout): {e}')
      time.sleep(RETRY_INTERVAL_SEC)

  return ssh_client


def parse_key_name(keyname) -> List[str]:
  """keyname => resource, username"""
  # Relies on resource name not containing -, validated in
  # validate_resource_name
  toks = keyname.split('-')
  if len(toks) != 2:
    return []  # some other keyname not launched by nexus
  else:
    return toks


###################
# Name validation:
# task name is used in machine name so must match AWS resource naming restrictions.
# run name is used in EFS directory name, so use a subset of Posix directory naming requirements. No . because that's used by ncluster naming convention
# prefix is used for both AWS resources and Unix file names so must match both
###################

aws_name_regexp_str = re.compile('^[a-zA-Z0-9+-=._:/@]*$')
aws_name_regexp = re.compile(aws_name_regexp_str)


def validate_aws_name(name) -> None:
  """Validate resource name using AWS name restrictions from # http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/Using_Tags.html#tag-restrictions"""
  assert len(name) <= 127
  # disallow unicode characters to avoid pain
  assert name == name.encode('ascii').decode('ascii'), f"Non-ascii chars found in '{name}"
  assert aws_name_regexp.match(
    name), f"Creating AWS resource with illegal characters, '{name}' must match regexp '{aws_name_regexp_str}'"


def validate_task_name(name):
  assert name == name.encode('ascii').decode('ascii'), f"Non-ascii chars found in '{name}'"
  # assert '.' not in name, f"can't use . in task name '{name}'"
  assert aws_name_regexp.match(name), f"Task name '{name}' must match regexp '{aws_name_regexp_str}'"


posix_regexp_str = '^[-_a-z0-9]+$'
posix_regexp = re.compile(posix_regexp_str)


def validate_prefix(name):
  """Check that name is valid as substitute for default prefix. Since it's used in unix filenames, key names, be more conservative than AWS requirements, just allow 30 chars, lowercase only."""
  assert len(name) <= 30
  assert posix_regexp.match(name), f"Prefix '{name}' must match '{posix_regexp_str}'"
  validate_aws_name(name)


def validate_run_name(name):
  """Name used for run. Used as part of instance name, tmux session name."""
  assert len(name) <= 30
  assert name == name.encode('ascii').decode('ascii'), f"Non-ascii chars found in '{name}"
  if name:  # allow empty run name
    assert posix_regexp.match(name), f"Run name '{name}' must match '{posix_regexp_str}'"


def create_name_tags(name):
  """Returns [{'Key': 'Name', 'Value': name}] """
  return [{'Key': 'Name', 'Value': name}]


def create_efs(name) -> str:
  efs_client = get_efs_client()
  token = str(int(time.time() * 1e6))  # epoch usec

  response = efs_client.create_file_system(CreationToken=token,
                                           PerformanceMode='generalPurpose')
  assert is_good_response(response)
  start_time = time.time()
  while True:
    try:

      response = efs_client.create_file_system(CreationToken=token,
                                               PerformanceMode='generalPurpose')
      assert is_good_response(response)
      time.sleep(RETRY_INTERVAL_SEC)
    except Exception as e:
      if 'FileSystemAlreadyExists' in str(e):
        break
      if response['Error']['Code'] == 'FileSystemAlreadyExists':
        break
      else:
        util.log_error(e)
      break

    if time.time() - start_time - RETRY_INTERVAL_SEC > RETRY_TIMEOUT_SEC:
      assert False, "Timeout exceeded creating EFS %s (%s)" % (token, name)

    time.sleep(RETRY_INTERVAL_SEC)

  # find efs id from given token
  response = efs_client.describe_file_systems()
  assert is_good_response(response)
  fs_id = extract_attr_for_match(response['FileSystems'], FileSystemId=-1,
                                 CreationToken=token)
  response = efs_client.create_tags(FileSystemId=fs_id,
                                    Tags=create_name_tags(name))
  assert is_good_response(response)

  # make sure EFS is now visible
  efs_dict = get_efs_dict()
  assert name in efs_dict
  return efs_dict[name]


def delete_efs_by_id(efs_id):
  """Deletion sometimes fails, try several times."""
  start_time = time.time()
  efs_client = get_efs_client()
  sys.stdout.write("deleting %s ... " % (efs_id,))
  while True:
    try:
      response = efs_client.delete_file_system(FileSystemId=efs_id)
      if is_good_response(response):
        print("succeeded")
        break
      time.sleep(RETRY_INTERVAL_SEC)
    except Exception as e:
      print("Failed with %s" % (e,))
      if time.time() - start_time - RETRY_INTERVAL_SEC < RETRY_TIMEOUT_SEC:
        print("Retrying in %s sec" % (RETRY_INTERVAL_SEC,))
        time.sleep(RETRY_INTERVAL_SEC)
      else:
        print("Giving up")
        break


def extract_attr_for_match(items, **kwargs):
  """Helper method to get attribute value for an item matching some criterion.
  Specify target criteria value as dict, with target attribute having value -1

  Example:
    to extract state of vpc matching given vpc id

  response = [{'State': 'available', 'VpcId': 'vpc-2bb1584c'}]
  extract_attr_for_match(response, State=-1, VpcId='vpc-2bb1584c') #=> 'available'"""

  # find the value of attribute to return
  query_arg = None
  for arg, value in kwargs.items():
    if value == -1:
      assert query_arg is None, "Only single query arg (-1 valued) is allowed"
      query_arg = arg
  result = []

  filterset = set(kwargs.keys())
  for item in items:
    match = True
    assert filterset.issubset(
      item.keys()), "Filter set contained %s which was not in record %s" % (
      filterset.difference(item.keys()),
      item)
    for arg in item:
      if arg == query_arg:
        continue
      if arg in kwargs:
        if item[arg] != kwargs[arg]:
          match = False
          break
    if match:
      result.append(item[query_arg])
  assert len(result) <= 1, "%d values matched %s, only allow 1" % (
    len(result), kwargs)
  if result:
    return result[0]
  return None


def get_tags(instance):
  """Returns instance tags."""

  return get_instance_property(instance, 'tags')


def get_public_ip(instance):
  return get_instance_property(instance, 'public_ip_address')


def get_ip(instance):
  return get_instance_property(instance, 'private_ip_address')


def get_instance_property(instance, property_name):
  """Retrieves property of an instance, keeps retrying until getting a non-None"""

  name = get_name(instance)
  value = None
  start_time = time.time()
  while True:
    if time.time() - start_time > RETRY_TIMEOUT_SEC:
      assert False, f"Timeout {RETRY_TIMEOUT_SEC} exceeded quering {get_instance_property} on {instance}"
    try:
      value = getattr(instance, property_name)
      if value is not None:
        break
      print(f"retrieving {property_name} on {name} produced None, retrying")
      time.sleep(RETRY_INTERVAL_SEC)
      instance.reload()
      continue
    except Exception as e:
      print(f"retrieving {property_name} on {name} failed with {e}, retrying")
      time.sleep(RETRY_INTERVAL_SEC)
      try:
        instance.reload()
      except Exception:
        pass
      continue

  return value


def call_with_retries(method, debug_string='',
                      retry_interval_sec=RETRY_INTERVAL_SEC,
                      retry_timeout_sec=RETRY_TIMEOUT_SEC,
                      **kwargs):
  start_time = time.time()
  value = None
  while True:
    if time.time() - start_time > retry_timeout_sec:
      assert False, f"Timeout {retry_timeout_sec} exceeded calling {method.__name__}"
    try:
      value = method(**kwargs)
      assert value is not None, f"{debug_string} was None"
      break
    except Exception as e:
      print(f"{debug_string} failed with {e.__class__}({e}), retrying for another {int(retry_timeout_sec - (time.time() - start_time))} seconds")
      time.sleep(retry_interval_sec)
      continue

  return value


def get_ec2_resource() -> EC2_ServiceResource:
  try:
    client = get_session().resource('ec2')
  except Exception as e:
    print(f"Failed with error '{e}'")
    print("To specify Virginia region, do 'export AWS_DEFAULT_REGION=us-east-1'")
    sys.exit()
  return client


def get_ec2_client() -> EC2_Client:
  try:
    client = get_session().client('ec2')
  except Exception as e:
    print(f"Failed with error '{e}'")
    print("To specify Virginia region, do 'export AWS_DEFAULT_REGION=us-east-1'")
    sys.exit()
  return client


def get_efs_client():
  while True:
    try:
      return get_session().client('efs')
    except Exception as e:
      # can get following
      # botocore.exceptions.DataNotFoundError: Unable to load data for: endpoints
      util.log(f"get_session().client('efs') failed with {e}, retrying")
      time.sleep(2)


def is_good_response(response):
  """Helper method to check if boto3 call was a success."""

  code = response["ResponseMetadata"]['HTTPStatusCode']
  # get response code 201 on EFS creation
  return 200 <= code < 300


def get_name(tags_or_instance_or_id):
  """Helper utility to extract name out of tags dictionary or intancce.
      [{'Key': 'Name', 'Value': 'nexus'}] -> 'nexus'
 
     Assert fails if there's more than one name.
     Returns '' if there's less than one name.
  """

  ec2 = get_ec2_resource()
  if hasattr(tags_or_instance_or_id, 'tags'):
    tags = tags_or_instance_or_id.tags
  elif isinstance(tags_or_instance_or_id, str):
    tags = ec2.Instance(tags_or_instance_or_id).tags
  elif tags_or_instance_or_id is None:
    return EMPTY_NAME
  else:
    assert isinstance(tags_or_instance_or_id,
                      Iterable), "expected iterable of tags"
    tags = tags_or_instance_or_id

  if not tags:
    return EMPTY_NAME
  names = [entry['Value'] for entry in tags if entry['Key'] == 'Name']
  if not names:
    return ''
  if len(names) > 1:
    assert False, "have more than one name: " + str(names)
  return names[0]


def wait_until_available(resource):
  """Waits until interval state becomes 'available'"""
  while True:
    resource.load()
    if resource.state == 'available':
      break
    time.sleep(RETRY_INTERVAL_SEC)


def maybe_create_placement_group(name='', max_retries=10):
  """Creates placement_group group or reuses existing one. Crash if unable to create
    placement_group group. If name is empty, ignores request."""

  if not name:
    return

  client = get_ec2_client()
  while True:
    try:
      client.describe_placement_groups(GroupNames=[name])
      print("Reusing placement_group group: " + name)
      break  # no Exception means group name was found
    except Exception:
      print("Creating placement_group group: " + name)
      try:
        _response = client.create_placement_group(GroupName=name,
                                                  Strategy='cluster')
      except botocore.exceptions.ClientError as e:
        if 'PlacementGroupLimitExceeded' in str(e):
          print(e)
          assert False, "Clean-up placement groups using 'ncluster cleanup_placement_groups'"
        print(f"Placement group creation failed with {e}, retrying")
        if e.response['Error']['Code'] == 'InvalidPlacementGroup.Duplicate':
          util.log(f"possible race condition, group {name} already created")
        pass

  counter = 0
  while True:
    try:
      res = client.describe_placement_groups(GroupNames=[name])
      res_entry = res['PlacementGroups'][0]
      if res_entry['State'] == 'available':
        assert res_entry['Strategy'] == 'cluster'
        break
    except Exception as e:
      print("Got exception: %s" % (e,))
    counter += 1
    if counter >= max_retries:
      assert False, f'Failed to create placement_group group {name} in {max_retries} attempts'
    time.sleep(RETRY_INTERVAL_SEC)


# https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/placement-groups.html#concepts-placement-groups
def instance_supports_placement_groups(instance_type: str):
  regex = re.compile(
    "^(m4|m5|m5d|c3|c4|c5|c5d|cc2.8xlarge|cr1.8xlarge|r3|r4|r5|r5d|x1|x1e|z1d|d2|h1|hs1.8xlarge|i2|i3|i3.metal|f1|g2|g3|p2|p3).*$",
    re.IGNORECASE)
  return regex.match(instance_type)


def instance_supports_efa(instance_type: str) -> bool:
  """Checks if instance supports Amazon Elastic Fabric Adapter"""
  # https://docs.aws.amazon.com/en_us/AWSEC2/latest/UserGuide/efa-start.html
  return instance_type in ['c5n.18xlarge', 'i3en.24xlarge', 'p3dn.24xlarge']


def create_spot_instances(launch_specs, spot_price=26, expiration_mins=15):
  """
    args:
      spot_price: default is $26 which is right above p3.16xlarge on demand price
      expiration_mins: this request only valid for this many mins from now
    """
  ec2c = get_ec2_client()

  num_tasks = launch_specs['MinCount'] or 1
  if 'MinCount' in launch_specs:
    del launch_specs['MinCount']
  if 'MaxCount' in launch_specs:
    del launch_specs['MaxCount']
  tags = None
  if 'TagSpecifications' in launch_specs:
    try:
      tags = launch_specs['TagSpecifications'][0]['Tags']
    except Exception:  # Tags entry not present
      pass
    del launch_specs['TagSpecifications']

  import pytz  # datetime is not timezone aware, use pytz to fix
  import datetime as dt
  now = dt.datetime.utcnow().replace(tzinfo=pytz.utc)

  spot_args = {}
  spot_args['LaunchSpecification'] = launch_specs
  spot_args['SpotPrice'] = str(spot_price)
  spot_args['InstanceCount'] = num_tasks
  spot_args['ValidUntil'] = now + dt.timedelta(minutes=expiration_mins)

  try:
    spot_requests = ec2c.request_spot_instances(**spot_args)
  except Exception as e:
    assert False, f"Spot instance request failed (out of capacity?), error was {e}"

  spot_requests = spot_requests['SpotInstanceRequests']
  instance_ids = wait_on_fulfillment(ec2c, spot_requests)

  print('Instances fullfilled...')
  ec2 = get_ec2_resource()
  instances = list(ec2.instances.filter(Filters=[{'Name': 'instance-id', 'Values': list(filter(None, instance_ids))}]))

  if not all(instance_ids):
    for i in instances:
      i.terminate()
    raise RuntimeError('Failed to create spot instances:', instance_ids)

  if tags:
    for i in instances:
      i.create_tags(Tags=tags)

  return instances


def wait_on_fulfillment(ec2c, reqs):
  def get_instance_id(req):
    while req['State'] != 'active':
      print('Waiting on spot fullfillment...')
      time.sleep(5)
      reqs_ = ec2c.describe_spot_instance_requests(
        Filters=[{'Name': 'spot-instance-request-id', 'Values': [req['SpotInstanceRequestId']]}])
      if not reqs_['SpotInstanceRequests']:
        print(f"SpotInstanceRequest for {req['SpotInstanceRequestId']} not found")
        continue
      req = reqs_['SpotInstanceRequests'][0]
      req_status = req['Status']
      if req_status['Code'] not in ['pending-evaluation', 'pending-fulfillment', 'fulfilled']:
        print('Spot instance request failed:', req_status['Message'])
        print('Cancelling request. Please try again or use on demand.')
        ec2c.cancel_spot_instance_requests(SpotInstanceRequestIds=[req['SpotInstanceRequestId']])
        print(req)
        return None
    instance_id = req['InstanceId']
    print('Fulfillment completed. InstanceId:', instance_id)
    return instance_id

  return [get_instance_id(req) for req in reqs]


def assert_zone_specific_config():
  """Make sure user specified zone"""

  assert get_zone(), f"EFA requires launching into specific zone, but zone not specified, try NCLUSTER_ZONE={get_region()}a"


def get_security_group_names() -> List[str]:
  """Returns all security group names associated with given ncluster configuration."""
  return [get_security_group_name(), get_security_group_nd_name()]


def get_efs_mount_command():
  region = get_region()
  efs_id = get_efs_dict()[get_prefix()]
  dns = f"{efs_id}.efs.{region}.amazonaws.com"
  cmd = f'sudo mkdir -p /ncluster && sudo mount -t nfs -o nfsvers=4.1,rsize=1048576,wsize=1048576,hard,timeo=600,retrans=2 {dns}:/ /ncluster'
  return cmd


def validate_local_keypair():
  key_name = get_keypair_name()
  keypair_fn = get_keypair_fn()
  if key_name in get_keypair_dict():
    assert os.path.exists(keypair_fn), f"Keypair '{key_name}' exists, but couldn't find corresponding .pem file under '{keypair_fn}'. Either get this file or delete it using 'ncluster fixkeys'"


def running_on_aws() -> bool:
  """
  Returns: True if running on AWS, False otherwise

  """

  # follow recipe from https://serverfault.com/a/575422/88745
  cmd = 'wget -q http://instance-data.ec2.internal && echo yes || echo no'
  p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
  (stdout, stderr) = p.communicate()
  if stderr:
    print(f"Warning, running_on_aws check got stderr '{stderr}'")
    return stdout.decode('ascii')

  if stdout == 'yes':
    return True
  if stdout == 'no':
    return False
  assert False, f"running_on_aws check got '{stdout}' which is neither 'yes' nor 'no'"




#!/usr/bin/env python
#
# Creates resources.
# To run standalone:
# python -m ncluster.ncluster_cloud_setup
#
# This script creates VPC/security group/keypair if not already present

import os
import sys
import time
from typing import Tuple, Any, Optional

from boto3_type_annotations.ec2 import SecurityGroup

from ncluster import aws_util as u
from ncluster import util

DRYRUN = False
DEBUG = True

# Names of Amazon resources that are created. These settings are fixed across
# all runs, and correspond to resources created once per user per region.

NFS_PORT = 2049
PUBLIC_TCP_RANGES = [
  22,  # ssh
  NFS_PORT,  # NFS port NFS peering between security groups
  # ipython notebook ports
  (8888, 8899),
  # redis port
  6379,
  # tensorboard ports
  (6006, 6016)
]

PUBLIC_UDP_RANGES = [NFS_PORT, (60000, 61000)]  # mosh ports


def network_setup() -> Tuple[Any, Any]:
  """Creates VPC if it doesn't already exists, configures it for public
  internet access, returns vpc, subnet, security_group"""

  # from https://gist.github.com/nguyendv/8cfd92fc8ed32ebb78e366f44c2daea6

  ec2 = u.get_ec2_resource()
  client = u.get_ec2_client()
  existing_vpcs = u.get_vpc_dict()
  zones = u.get_zones()

  # create VPC from scratch. Remove this if default VPC works well enough.
  create_non_default_vpc = False

  if create_non_default_vpc:
    vpc_name = u.get_vpc_name()
    if u.get_vpc_name() in existing_vpcs:
      print("Reusing VPC " + vpc_name)
      vpc = existing_vpcs[vpc_name]
      subnets = list(vpc.subnets.all())
      assert len(subnets) == len(
        zones), "Has %s subnets, but %s zones, something went wrong during resource creation, try delete_resources.py/create_resources.py" % (
        len(subnets), len(zones))

    else:
      print("Creating VPC " + vpc_name)
      vpc = ec2.create_vpc(CidrBlock='192.168.0.0/16')

      # enable DNS on the VPC
      response = vpc.modify_attribute(EnableDnsHostnames={"Value": True})
      assert u.is_good_response(response)
      response = vpc.modify_attribute(EnableDnsSupport={"Value": True})
      assert u.is_good_response(response)

      vpc.create_tags(Tags=u.create_name_tags(vpc_name))
      vpc.wait_until_available()

    gateways = u.get_gateway_dict(vpc)
    gateway_name = u.get_gateway_name()
    if gateway_name in gateways:
      print("Reusing gateways " + gateway_name)
    else:
      print("Creating internet gateway " + gateway_name)
      ig = ec2.create_internet_gateway()
      ig.attach_to_vpc(VpcId=vpc.id)
      ig.create_tags(Tags=u.create_name_tags(gateway_name))

      # check that attachment succeeded
      attach_state = u.extract_attr_for_match(ig.attachments, State=-1,
                                              VpcId=vpc.id)
      assert attach_state == 'available', "vpc %s is in state %s" % (vpc.id,
                                                                     attach_state)
      route_table = vpc.create_route_table()
      route_table_name = u.get_route_table_name()
      route_table.create_tags(Tags=u.create_name_tags(route_table_name))

      dest_cidr = '0.0.0.0/0'
      route_table.create_route(
        DestinationCidrBlock=dest_cidr,
        GatewayId=ig.id
      )
      # check success
      for route in route_table.routes:
        # result looks like this
        # ec2.Route(route_table_id='rtb-a8b438cf',
        #    destination_cidr_block='0.0.0.0/0')
        if route.destination_cidr_block == dest_cidr:
          break
      else:
        # sometimes get
        #      AssertionError: Route for 0.0.0.0/0 not found in [ec2.Route(route_table_id='rtb-cd9153b0', destination_cidr_block='192.168.0.0/16')]
        # TODO: add a wait/retry?
        assert False, "Route for %s not found in %s" % (dest_cidr,
                                                        route_table.routes)

      assert len(zones) <= 16  # for cidr/20 to fit into cidr/16
      ip = 0
      for zone in zones:
        cidr_block = '192.168.%d.0/20' % (ip,)
        ip += 16
        print("Creating subnet %s in zone %s" % (cidr_block, zone))
        subnet = vpc.create_subnet(CidrBlock=cidr_block,
                                   AvailabilityZone=zone)
        subnet.create_tags(Tags=[{'Key': 'Name', 'Value': f'{vpc_name}-subnet'},
                                 {'Key': 'Region', 'Value': zone}])
        response = client.modify_subnet_attribute(
          MapPublicIpOnLaunch={'Value': True},
          SubnetId=subnet.id
        )
        assert u.is_good_response(response)
        u.wait_until_available(subnet)
        assert subnet.map_public_ip_on_launch, "Subnet doesn't enable public IP by default, why?"

        route_table.associate_with_subnet(SubnetId=subnet.id)

  #  Setup security group for non-default VPC
  #  existing_security_groups = u.get_security_group_dict()
  #  security_group_nd_name = u.get_security_group_nd_name()
  # if security_group_nd_name in existing_security_groups:
  #   print("Reusing non-default security group " + security_group_nd_name)
  #   security_group_nd = existing_security_groups[security_group_nd_name]
  #   assert security_group_nd.vpc_id == vpc.id, f"Found non-default security group {security_group_nd} " \
  #                                              f"attached to {security_group_nd.vpc_id} but expected {vpc.id}"
  # else:
  #   security_group_nd = create_security_group(security_group_nd_name, vpc.id)

  # Setup things on default VPC for zone-agnostic launching
  vpc = u.get_default_vpc()
  if not vpc:
    util.log(f"Creating default VPC for region {u.get_region()}")
    client.create_default_vpc()
  vpc = u.get_default_vpc()
  assert vpc, "Could not create default VPC?"

  existing_security_groups = u.get_security_group_dict()
  security_group_name = u.get_security_group_name()
  if security_group_name in existing_security_groups:
    print("Reusing security group " + security_group_name)
    security_group = existing_security_groups[security_group_name]
    assert security_group.vpc_id == vpc.id, f"Found security group {security_group} " \
                                            f"attached to {security_group.vpc_id} but expected {vpc.id}"
  else:
    security_group = create_security_group(security_group_name, vpc.id)
    #  Uncomment the following when setting up two VPC's
    #  security_group = create_security_group(security_group_name, vpc.id, security_group_nd)

  return vpc, security_group


def keypair_setup():
  """Creates keypair if necessary, saves private key locally, returns contents
  of private key file."""

  os.system('mkdir -p ' + u.PRIVATE_KEY_LOCATION)
  keypair_name = u.get_keypair_name()
  keypair = u.get_keypair_dict().get(keypair_name, None)
  keypair_fn = u.get_keypair_fn()
  if keypair:
    print("Reusing keypair " + keypair_name)
    # check that local pem file exists and is readable
    assert os.path.exists(
      keypair_fn), "Keypair %s exists, but corresponding .pem file %s is not found, delete keypair %s through console and run again to recreate keypair/.pem together" % (
      keypair_name, keypair_fn, keypair_name)
    keypair_contents = open(keypair_fn).read()
    assert len(keypair_contents) > 0
  else:
    print("Creating keypair " + keypair_name)
    ec2 = u.get_ec2_resource()
    assert not os.path.exists(
      keypair_fn), "previous keypair exists, delete it with 'sudo rm %s' and also delete corresponding keypair through console" % (
      keypair_fn)
    keypair = ec2.create_key_pair(KeyName=keypair_name)

    open(keypair_fn, 'w').write(keypair.key_material)
    os.system('chmod 400 ' + keypair_fn)

  return keypair


def placement_group_setup(group_name):
  """Creates placement_group group if necessary. Returns True if new placement_group
  group was created, False otherwise."""

  existing_placement_groups = u.get_placement_group_dict()

  group = existing_placement_groups.get(group_name, None)
  if group:
    assert group.state == 'available'
    assert group.strategy == 'cluster'
    print("Reusing group ", group.name)
    return group

  print("Creating group " + group_name)
  ec2 = u.get_ec2_resource()
  group = ec2.create_placement_group(GroupName=group_name, Strategy='cluster')
  return group


def create_security_group(security_group_name: str, vpc_id: str, other_group: Optional[SecurityGroup] = None):
  """Creates security group with proper ports open. Optionally allows all traffic from other_group"""
  print("Creating security group " + security_group_name)
  ec2 = u.get_ec2_resource()

  security_group: SecurityGroup = ec2.create_security_group(
    GroupName=security_group_name, Description=security_group_name,
    VpcId=vpc_id)

  security_group.create_tags(Tags=u.create_name_tags(security_group_name))

  # allow ICMP access for public ping
  security_group.authorize_ingress(
    CidrIp='0.0.0.0/0',
    IpProtocol='icmp',
    FromPort=-1,
    ToPort=-1
  )

  # open public ports
  # always include SSH port which is required for basic functionality
  assert 22 in PUBLIC_TCP_RANGES, "Must enable SSH access"
  for port in PUBLIC_TCP_RANGES:
    if util.is_iterable(port):
      assert len(port) == 2
      from_port, to_port = port
    else:
      from_port, to_port = port, port

    response = security_group.authorize_ingress(IpProtocol="tcp",
                                                CidrIp="0.0.0.0/0",
                                                FromPort=from_port,
                                                ToPort=to_port)
    assert u.is_good_response(response)

  for port in PUBLIC_UDP_RANGES:
    if util.is_iterable(port):
      assert len(port) == 2
      from_port, to_port = port
    else:
      from_port, to_port = port, port

    response = security_group.authorize_ingress(IpProtocol="udp",
                                                CidrIp="0.0.0.0/0",
                                                FromPort=from_port,
                                                ToPort=to_port)
    assert u.is_good_response(response)

  def authorize_from_group(this_security_group: SecurityGroup, other_security_group: SecurityGroup):
    """Helper function to authorize all traffic from other_group. Can be used to authorized within-group traffic as
    authorize_from_group(group, group)"""

    # Authorizing ingress doesn't work with security group names in a non-default VPC,
    # so must use more complicated syntax: https://github.com/boto/boto3/issues/158
    response_ = {}
    for protocol in ['icmp']:
      try:
        rule = {'FromPort': -1,
                'IpProtocol': protocol,
                'IpRanges': [],
                'PrefixListIds': [],
                'ToPort': -1,
                'UserIdGroupPairs': [{'GroupId': other_security_group.id}]}
        response_ = this_security_group.authorize_ingress(IpPermissions=[rule])
      except Exception as e:
        if response_['Error']['Code'] == 'InvalidPermission.Duplicate':
          print("Warning, got " + str(e))
        else:
          assert False, "Failed while authorizing icml ingress with " + str(e)

    for protocol in ['tcp', 'udp']:
      try:
        rule = {'FromPort': 0,
                'IpProtocol': protocol,
                'IpRanges': [],
                'PrefixListIds': [],
                'ToPort': 65535,
                'UserIdGroupPairs': [{'GroupId': other_security_group.id}]}
        response_ = this_security_group.authorize_ingress(IpPermissions=[rule])
      except Exception as e:
        if response_['Error']['Code'] == 'InvalidPermission.Duplicate':
          print("Warning, got " + str(e))
        else:
          assert False, "Failed while authorizing tcp/udp ingress with " + str(e)

    # authorize EFA traffic
    user_id = u.get_account_number()
    response = None
    try:
      rule = {
         "IpProtocol": "-1",
         "Ipv6Ranges": [],
         "PrefixListIds": [],
         'UserIdGroupPairs': [{'Description': 'efa', 'GroupId': other_security_group.id, 'UserId': user_id}]
      }
      response_ = this_security_group.authorize_ingress(IpPermissions=[rule])
      assert u.is_good_response(response_), str(response)

      rule = {
        "IpProtocol": "-1",
        "PrefixListIds": [],
        'UserIdGroupPairs': [{'Description': 'efa',
                              'GroupId': other_security_group.id,
                              'UserId': user_id}]
      }
      response_ = this_security_group.authorize_egress(IpPermissions=[rule])
      assert u.is_good_response(response_), str(response)

    except Exception as e:
      if 'Error' in response_ and 'Code' in response['Error'] and response_['Error']['Code'] == 'InvalidPermission.Duplicate':
        print(f"Warning while authorizing ingress from {this_security_group.description} ({this_security_group.id}) to "
              f"{other_security_group.description} ({other_security_group.id}) with message '{e}'")

      else:
        assert False, (f"Failed while authorizing ingress from {this_security_group.description} ({this_security_group.id}) to "
                       f"{other_security_group.description} ({other_security_group.id}) with message '{e}' and response '{response}'")

  authorize_from_group(security_group, security_group)
  # if using multiple security groups, which is required for the case of default + non-default VPC
  # also authorize all traffic between them
  if other_group:
    authorize_from_group(security_group, other_group)
    authorize_from_group(other_group, security_group)

  return security_group


def create_resources():
  print(f"Creating {u.get_prefix()} resources in region {u.get_region()}")

  vpc, security_group = network_setup()
  keypair_setup()  # saves private key locally to keypair_fn

  # create EFS
  efss = u.get_efs_dict()
  efs_name = u.get_efs_name()
  efs_id = efss.get(efs_name, '')
  if not efs_id:
    print("Creating EFS " + efs_name)
    efs_id = u.create_efs(efs_name)
  else:
    print("Reusing EFS " + efs_name)

  efs_client = u.get_efs_client()

  # create mount target for each subnet in the VPC

  # added retries because efs is not immediately available
  max_failures = 10
  retry_interval_sec = 1
  for subnet in vpc.subnets.all():
    for retry_attempt in range(max_failures):
      try:
        sys.stdout.write(
          "Creating efs mount target for %s ... " % (subnet.availability_zone,))
        sys.stdout.flush()
        response = efs_client.create_mount_target(FileSystemId=efs_id,
                                                  SubnetId=subnet.id,
                                                  SecurityGroups=[
                                                    security_group.id])
        if u.is_good_response(response):
          print("success")
          break
      except Exception as e:
        if 'already exists' in str(e):  # ignore "already exists" errors
          print('already exists')
          break

        # Takes couple of seconds for EFS to come online, with
        # errors like this:
        # Creating efs mount target for us-east-1f ... Failed with An error occurred (IncorrectFileSystemLifeCycleState) when calling the CreateMountTarget operation: None, retrying in 1 sec

        print("Got %s, retrying in %s sec" % (str(e), retry_interval_sec))
        time.sleep(retry_interval_sec)
    else:
      print("Giving up.")


if __name__ == '__main__':
  create_resources()



import os

from . import aws_backend
from . import local_backend
from . import backend
from . import aws_util as u
import collections

from . import ncluster_globals

_backend: type(backend) = backend


def get_logdir_root() -> str:
  return ncluster_globals.LOGDIR_ROOT


def set_logdir_root(logdir_root):
  """Globally changes logdir root for all runs."""
  ncluster_globals.LOGDIR_ROOT = logdir_root


def set_backend(backend_name: str):
  """Sets backend (local or aws)"""
  global _backend, _backend_name
  _backend_name = backend_name

  assert not ncluster_globals.task_launched, "Not allowed to change backend after launching a task (this pattern is error-prone)"
  if backend_name == 'aws':
    _backend = aws_backend
  elif backend_name == 'local':
    _backend = local_backend
  else:
    assert False, f"Unknown backend {backend_name}"

  # take default value for logdir root from backend
  ncluster_globals.LOGDIR_ROOT = _backend.DEFAULT_LOGDIR_ROOT


def use_aws():
  set_backend('aws')


def use_local():
  set_backend('local')


def get_backend() -> str:
  """Returns backend name, ie "local" or "aws" """
  return _backend_name


def get_backend_module() -> backend:
  return _backend


def running_locally():
  return get_backend() == 'local'


def get_region() -> str:
  if _backend != local_backend:
    return u.get_region()
  else:
    return 'local'


def get_zone() -> str:
  if _backend != local_backend:
    return u.get_zone()
  else:
    return 'local'


#  def make_run(name='', **kwargs):
#  return _backend.Run(name, **kwargs)


# Use factory methods task=create_task instead of relying solely on constructors task=Task() because underlying hardware resources may be reused between instantiations
# For instance, one may create a Task initialized with an instance that was previous created for this kind of task
# Factory method will make the decision to recreate or reuse such resource, and wrap this resource with a Task object.
def make_task(name: str = '',
              run_name: str = '',
              install_script: str = '',
              **kwargs) -> backend.Task:
  return _backend.make_task(name=name, run_name=run_name,
                            install_script=install_script, **kwargs)


def make_job(name: str = '',
             run_name: str = '',
             num_tasks: int = 0,
             install_script: str = '',
             **kwargs
             ) -> backend.Job:
  """
  Create a job using current backend. Blocks until all tasks are up and initialized.

  Args:
    name: name of the job
    run_name: name of the run (auto-assigned if empty)
    num_tasks: number of tasks
    install_script: bash-runnable script
    **kwargs:

  Returns:
    backend.Job
  """
  return _backend.make_job(name=name, run_name=run_name, num_tasks=num_tasks,
                           install_script=install_script, **kwargs)


def make_run(name: str = '', **kwargs) -> backend.Run:
  return _backend.make_run(name=name, **kwargs)


def add_authorized_keys(keys: list) -> str:
  """Add given public keys to list of keys authorized to access ncluster instances."""

  current_keys = os.environ.get('NCLUSTER_AUTHORIZED_KEYS', '')
  if current_keys:
    current_keys_list = current_keys.split(';')
  else:
    current_keys_list = []

  current_keys_list.extend(keys)
  current_keys_str = ';'.join(current_keys_list)
  os.environ['NCLUSTER_AUTHORIZED_KEYS'] = current_keys_str
  return current_keys_str


# TODO: remove?
def join(things_to_join):
  if isinstance(things_to_join, collections.Iterable):
    for thing in things_to_join:
      thing.join()
  else:
    things_to_join.join()



"""Module that keeps global state of ncluster tasks, such as naming,
connection of tasks to runs

run refers to string name
run_object refers to Run object corresponding to that name

"""
import os
import sys
from typing import Dict, Any, List
from . import util
from . import backend

LOGDIR_ROOT = None
task_launched = False  # keep track whether anything has been launched

task_counter = 0
job_counter = 0
run_counter = 0

run_dict: Dict[str, Any] = {}
task_run_dict: Dict[backend.Task, str] = {}
run_task_dict: Dict[str, List[backend.Task]] = {}
run_logdir_dict: Dict[str, str] = {}

tasks_seen: List[backend.Task] = []  # list of all tasks created

enforce_placement_group_val = False


def enforce_placement_group():
  """Enforces all tasks to be launched into placement group."""
  global enforce_placement_group_val
  enforce_placement_group_val = True


def unenforce_placement_group():
  """Enforces all tasks to be launched into placement group."""
  global enforce_placement_group_val
  enforce_placement_group_val = False


def is_enforced_placement_group():
  return enforce_placement_group_val


def auto_assign_task_name_if_needed(name, instance_type='', image_name='',
                                    tasks=1):
  global task_counter
  if name:
    return name

  main_script = os.path.abspath(sys.argv[0])
  script_id = util.alphanumeric_hash(
    f"{main_script}-{instance_type}-{image_name}-{tasks}")
  name = f"unnamedtask-{task_counter}-{script_id}"
  task_counter += 1
  return name


def auto_assign_job_name_if_needed(name):
  global job_counter
  if name:
    return name
  script_id = util.alphanumeric_hash(sys.argv[0])
  name = f"unnamedjob-{job_counter}-{script_id}"
  job_counter += 1
  return name


def auto_assign_run_name_if_needed(name):
  global run_counter
  if name:
    return name
  script_id = util.alphanumeric_hash(sys.argv[0])
  name = f"unnamedrun-{run_counter}-{script_id}"
  run_counter += 1
  return name


# def add_job_to_run(job, run_name):
#   global run_dict, job_run_dict
#   return job_run_dict.get(job, '')
#

# def register_run(name: str, run):
#   global run_dict, placement_dict
#   run_dict[name] = run
#


def register_task(task: Any, run_name: str):
  global task_run_dict, run_task_dict, tasks_seen
  assert task.name not in tasks_seen
  tasks_seen.append(task.name)
  task_run_dict[task] = run_name
  run_task_dict.setdefault(run_name, []).append(task)


def register_run(run: backend.Run, run_name: str) -> None:
  print(f"Registering run {run_name}")
  assert run_name not in run_dict
  assert run_name  # empty name reserved to mean no run
  run_dict[run_name] = run


def is_chief(task: backend.Task, run_name: str):
  """Returns True if task is chief task in the corresponding run"""
  global run_task_dict
  if run_name not in run_task_dict:
    return True
  task_list = run_task_dict[run_name]
  assert task in task_list, f"Task {task.name} doesn't belong to run {run_name}"
  return task_list[0] == task


def get_chief(run_name: str):
  assert run_name in run_task_dict, f"Run {run_name} doesn't exist"
  tasks = run_task_dict[run_name]
  assert tasks, f"Run {run_name} had tasks {tasks}, expected non-empty list"
  return tasks[0]


def get_logdir(run_name: str):
  """Returns logdir for this run. It is the job of logdir creator to set logdir for this run"""

  if not run_name:
    return '/tmp'
  return run_logdir_dict.get(run_name, '')


def set_logdir(run_name, logdir):
  assert run_name not in run_logdir_dict, f"logdir for run {run_name} has already been set to {run_logdir_dict[run_name]}, trying to change it to {logdir} is illegal"
  run_logdir_dict[run_name] = logdir


def get_run_for_task(task: backend.Task) -> str:
  """Gets run name associated with given Task"""
  return task_run_dict.get(task, '')


def get_run_object(run_name: str) -> backend.Run:
  return run_dict.get(run_name, None)


def create_run_if_needed(run_name, run_creation_callback) -> backend.Run:
  if run_name in run_dict:
    return run_dict[run_name]
  run = run_creation_callback(run_name)
  return run


_should_skip_setup = False


def set_should_skip_setup(val):
  global _should_skip_setup
  if val:
    util.log("skipping setup for all subsequent tasks/jobs")
  _should_skip_setup = val


def should_skip_setup():
  return _should_skip_setup



"""
Various helper utilities used internally by ncluster project, that are not explicitly tied to AWS
"""

import os
import random
import string
import sys
import time
from collections import Iterable
import shlex

from typing import Optional

import portalocker

# starting value for now_micros (Aug 31, 2018), using this to make various timestamped names shorter
EPOCH_MICROS = 1535753974788163


# Whitelist of temporary settings customizable through env vars. These are work-arounds for issues that are
# don't have permanent solutions yet. Keep this small to avoid many moving parts.
env_settings = {
  'NCLUSTER_AUTHORIZED_KEYS',         # public keys used to authorize ssh access on all instances
  'NCLUSTER_AWS_FAST_ROOTDISK',       # request $1/hour high performance AWS disk
  'NCLUSTER_AWS_PLACEMENT_GROUP',     # name of placement group to use, use when adding machines to a previous launched job
  'NCLUSTER_DISABLE_PDB_HANDLER',     # don't intercept pdb exception by default
  #  'NCLUSTER_IMAGE',
  'NCLUSTER_SSH_USERNAME',            # used as workaround when Amazon Linux detection fails
  'NCLUSTER_ZONE',                    # zone spec for when automatic zone fails (p3dn's + spot instances)
  'NCLUSTER_AWS_FORCE_CREATE_RESOURCES',  # AWS resources are created ignoring automatic existence checks
}


# keep this here instead of aws_backend because it's used by its dependency aws_util
VALID_REGIONS = ['us-east-2',
                 'us-east-1',
                 'us-west-1',       # An error occurred (Unsupported) when calling the RunInstances operation
                 'us-west-2',
                 'ap-east-1',       # doesn't have ec2
                 'ap-south-1',      # no EFS
                 'ap-northeast-3',  # An error occurred (OptInRequired) when calling the DescribeVpcs operation
                 'ap-northeast-2',
                 'ap-southeast-1',
                 'ap-southeast-2',
                 'ap-northeast-1',
                 'ca-central-1',
                 'cn-north-1',      # account number
                 'cn-northwest-1',  # account number
                 'eu-central-1',
                 'eu-west-1',
                 'eu-west-2',
                 'eu-west-3',       # no EFS
                 'eu-north-1',      # no EFS
                 'sa-east-1',       # no EFS
                 'us-gov-east-1',   # not authorized
                 'us-gov-west-1',   # not authorized
                 ]

# print/validate custom settings
for v in os.environ:
  if v.startswith('NCLUSTER'):
    assert v in env_settings, f"Custom setting '{v}'='{os.environ[v]}' not in settings whitelist, if you" \
      f"are sure you need this setting, add it to the env_settings in {os.path.basename(__file__)}, otherwise 'unset {v}'"
    # the following are often set by default, so don't print them
    if v in {'NCLUSTER_AUTHORIZED_KEYS', 'NCLUSTER_ZONE'}:
      continue

    sys.stderr.write(f"ncluster env setting {v}={os.environ[v]}\n")


def is_iterable(k):
  return isinstance(k, Iterable)


def now_micros(absolute=False) -> int:
  """Return current micros since epoch as integer."""
  micros = int(time.time() * 1e6)
  if absolute:
    return micros
  return micros - EPOCH_MICROS


def now_millis(absolute=False) -> int:
  """Return current millis since epoch as integer."""
  millis = int(time.time() * 1e3)
  if absolute:
    return millis
  return millis - EPOCH_MICROS // 1000


def current_timestamp() -> str:
  # timestamp format from https://github.com/tensorflow/tensorflow/blob/155b45698a40a12d4fef4701275ecce07c3bb01a/tensorflow/core/platform/default/logging.cc#L80
  current_seconds = time.time()
  remainder_micros = int(1e6 * (current_seconds - int(current_seconds)))
  time_str = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(current_seconds))
  full_time_str = "%s.%06d" % (time_str, remainder_micros)
  return full_time_str


def log_error(*args, **kwargs):
  print(f"Error encountered {args} {kwargs}")


def log(*args, **kwargs):
  print(f"{args} {kwargs}")


def install_pdb_handler():
  """Automatically start pdb:
      1. CTRL+\\ breaks into pdb.
      2. pdb gets launched on exception.
  """

  import signal
  import pdb

  def handler(_signum, _frame):
    pdb.set_trace()
  signal.signal(signal.SIGQUIT, handler)

  # Drop into PDB on exception
  # from https://stackoverflow.com/questions/13174412
  def info(type_, value, tb):
   if hasattr(sys, 'ps1') or not sys.stderr.isatty():
      # we are in interactive mode or we don't have a tty-like
      # device, so we call the default hook
      sys.__excepthook__(type_, value, tb)
   else:
      import traceback
      import pdb
      # we are NOT in interactive mode, print the exception...
      traceback.print_exception(type_, value, tb)
      print()
      # ...then start the debugger in post-mortem mode.
      pdb.pm()

  sys.excepthook = info


def shell_add_echo(script):
  """Goes over each line script, adds "echo cmd" in front of each cmd.

  ls a

  becomes

  echo * ls a
  ls a
  """
  new_script = ""
  for cmd in script.split('\n'):
    cmd = cmd.strip()
    if not cmd:
      continue
    new_script += "echo \\* " + shlex.quote(cmd) + "\n"
    new_script += cmd + "\n"
  return new_script


def shell_strip_comment(cmd):
  """ hi # testing => hi"""
  if '#' in cmd:
    return cmd.split('#', 1)[0]
  else:
    return cmd


def random_id(k=5):
  """Random id to use for AWS identifiers."""
  #  https://stackoverflow.com/questions/2257441/random-string-generation-with-upper-case-letters-and-digits-in-python
  return ''.join(random.choices(string.ascii_lowercase + string.digits, k=k))


def alphanumeric_hash(s: str, size=5):
  """Short alphanumeric string derived from hash of given string"""
  import hashlib
  import base64
  hash_object = hashlib.md5(s.encode('ascii'))
  s = base64.b32encode(hash_object.digest())
  result = s[:size].decode('ascii').lower()
  return result


def reverse_taskname(name: str) -> str:
  """
  Reverses components in the name of task. Reversed convention is used for filenames since
  it groups log/scratch files of related tasks together

  0.somejob.somerun -> somerun.somejob.0
  0.somejob -> somejob.0
  somename -> somename

  Args:
    name: name of task

  """
  components = name.split('.')
  assert len(components) <= 3
  return '.'.join(components[::-1])


def is_bash_builtin(cmd):
  """Return true if command is invoking bash built-in
  """
  # from compgen -b
  bash_builtins = ['alias', 'bg', 'bind', 'alias', 'bg', 'bind', 'break',
                   'builtin', 'caller', 'cd', 'command', 'compgen', 'complete',
                   'compopt', 'continue', 'declare', 'dirs', 'disown', 'echo',
                   'enable', 'eval', 'exec', 'exit', 'export', 'false', 'fc',
                   'fg', 'getopts', 'hash', 'help', 'history', 'jobs', 'kill',
                   'let', 'local', 'logout', 'mapfile', 'popd', 'printf',
                   'pushd', 'pwd', 'read', 'readarray', 'readonly', 'return',
                   'set', 'shift', 'shopt', 'source', 'suspend', 'test',
                   'times', 'trap', 'true', 'type', 'typeset', 'ulimit',
                   'umask', 'unalias', 'unset', 'wait']
  toks = cmd.split()
  if toks and toks[0] in bash_builtins:
    return True
  return False


def is_set(name: str) -> bool:
  """Helper method to check if given property is set"""
  assert name in env_settings

  val = os.environ.get(name, '0')
  assert val == '0' or val == '1', f"env var {name} has value {val}, expected 0 or 1"
  return val == '1'


def get_env(name: str) -> Optional[str]:
  """Helper method to retrieve custom env setting, returns None if not set"""
  assert name in env_settings
  return os.environ.get(name, None)


def set_env(name: str, value: str) -> None:
  """Helper method to set custom env setting"""
  assert name in env_settings
  os.environ[name] = value


def assert_script_in_current_directory():
  """Assert fail if current directory is different from location of the script"""

  script = sys.argv[0]
  assert os.path.abspath(os.path.dirname(script)) == os.path.abspath(
    '.'), f"Change into directory of script {script} and run again."


def validate_ncluster_job_name(name):
  assert name.count(
    '.') <= 1, "Job name has too many .'s (see ncluster design: Run/Job/Task hierarchy for  convention)"


def toseconds(dt) -> float:
  """Converts datetime object to seconds."""
  return time.mktime(dt.utctimetuple())


def wait_for_file(fn: str, max_wait_sec: int = 60,
                  check_interval: float = 1) -> bool:
    """
    Waits for file maximum of max_wait_sec. Returns True if file was detected within specified max_wait_sec
    Args:
      fn: filename
      max_wait_sec: how long to wait in seconds
      check_interval: how often to check in seconds
    Returns:
      False if waiting was was cut short by max_wait_sec limit, True otherwise
    """
    log("Waiting for file", fn)
    start_time = time.time()
    while True:
      if time.time() - start_time > max_wait_sec:
        log(f"Timeout exceeded ({max_wait_sec} sec) for {fn}")
        return False
      if not os.path.exists(fn):
        time.sleep(check_interval)
        continue
      else:
        break
    return True


# locations of default keypair
ID_RSA = os.environ['HOME'] + '/.ssh/id_rsa'
ID_RSA_PUB = ID_RSA + '.pub'


def setup_local_ssh_keys() -> str:
  """Sanity checks on local ssh keypair and regenerate it if necessary. Returns location of public keypair file"""

  if os.path.exists(ID_RSA_PUB):
    assert os.path.exists(ID_RSA), f"Public key {ID_RSA_PUB} exists but private key {ID_RSA} not found, delete {ID_RSA_PUB} and run again to regenerate pair"
    log(f"Found local keypair {ID_RSA}")
  elif os.path.exists(ID_RSA):
    assert os.path.exists(ID_RSA_PUB), f"Private key {ID_RSA} exists but public key {ID_RSA_PUB} not found, delete {ID_RSA} and run again to regenerate pair"
    log(f"Found local keypair {ID_RSA}")
  else:
    log(f"Generating keypair {ID_RSA}")
    with portalocker.Lock(ID_RSA+'.lock', timeout=5) as _:
      os.system(f"ssh-keygen -t rsa -f {ID_RSA} -N ''")
    os.system(f'rm {ID_RSA}.lock')

  return ID_RSA_PUB


def get_authorized_keys() -> str:
  """Appends local public key to NCLUSTER_AUTHORIZED_KEYS and returns in format key1;key2;key3;
  The result can be assigned back to NCLUSTER_AUTHORIZED_KEYS env var"""

  assert os.path.exists(ID_RSA_PUB), f"{ID_RSA_PUB} not found, make sure to run 'ncluster keys'"

  current_key = open(ID_RSA_PUB).read().strip()
  auth_keys = os.environ.get('NCLUSTER_AUTHORIZED_KEYS', '')
  return auth_keys+';'+current_key+';'


def get_public_key() -> str:
  """Returns public key, creating it if needed"""

  if not os.path.exists(ID_RSA_PUB):
    print(f"{ID_RSA_PUB} not found, running sure to run setup_local_ssh_keys()")
    setup_local_ssh_keys()
    assert os.path.exists(ID_RSA_PUB)

  return open(ID_RSA_PUB).read().strip()



"""Interface for job launching backend.

Run/Job and Task are container classes encapsulating functionality.
User creates them through make_run/make_job/make_task methods

"""
# Job launcher Python API: https://docs.google.com/document/d/1yTkb4IPJXOUaEWksQPCH7q0sjqHgBf3f70cWzfoFboc/edit
# AWS job launcher (concepts): https://docs.google.com/document/d/1IbVn8_ckfVO3Z9gIiE0b9K3UrBRRiO9HYZvXSkPXGuw/edit
import threading
import time
from typing import List, Tuple, Any, Optional

from . import util

# aws_backend.py
# local_backend.py

LOGDIR_ROOT: Optional[str] = None  # location of logdir for this backend

"""
backend = aws_backend # alternatively, backend=tmux_backend to launch jobs locally in separate tmux sessions
run = backend.make_run("helloworld")  # sets up /efs/runs/helloworld
worker_job = run.make_job("worker", instance_type="g3.4xlarge", num_tasks=4, ami=ami, setup_script=setup_script)
ps_job = run.make_job("ps", instance_type="c5.xlarge", num_tasks=4, ami=ami, setup_script=setup_script)
setup_tf_config(worker_job, ps_job)
ps_job.run("python cifar10_main.py --num_gpus=0")  # runs command on each task
worker_job.run("python cifar10_main.py --num_gpus=4")

tb_job = run.make_job("tb", instance_type="m4.xlarge", num_tasks=1, public_port=6006)
tb_job.run("tensorboard --logdir=%s --port=%d" %(run.logdir, 6006))
# when job has one task, job.task[0].ip can be accessed as job.ip
print("See TensorBoard progress on %s:%d" %(tb_job.ip, 6006))
print("To interact with workers: %s" %(worker_job.connect_instructions))


To reconnect to existing job:

"""


class Task:
  name: str
  ip: Optional[str]
  public_ip: Optional[str]
  run_counter: int
  # location where temporary files from interfacing with task go locally
  local_scratch: Optional[str]
  # location where temporary files from interfacing with task go on task
  remote_scratch: Optional[str]
  job: Any  # can't declare Job because of circular dependency

  def __init__(self, name=''):
    """Wraps execution resources into a task. Runs install_script if present"""
    self.last_status = None
    self.name = name
    self.instance = None
    self.install_script = None
    self.job = None
    self.kwargs = None
    self.public_ip = None
    self.ip = None
    self.logdir_ = None

  @property
  def logdir(self):
    raise NotImplementedError()

  def run(self, cmd: str, non_blocking=False, ignore_errors=False):
    """Runs command on given task."""
    raise NotImplementedError()

  def run_with_output(self, cmd, non_blocking=False, ignore_errors=False) -> \
          Tuple[str, str]:
    """

    Args:
      cmd: single line shell command to run
      non_blocking (bool): if True, does not wait for command to finish
      ignore_errors: if True, will succeed even if command failed

    Returns:
      Contents of stdout/stderr as strings.
    Raises
      RuntimeException: if command produced non-0 returncode

    """

    assert '\n' not in cmd, "Do not support multi-line commands"
    cmd: str = cmd.strip()
    if not cmd or cmd.startswith('#'):  # ignore empty/commented out lines
      return '', ''

    stdout_fn = f"{self.remote_scratch}/{self.run_counter+1}.stdout"
    stderr_fn = f"{self.remote_scratch}/{self.run_counter+1}.stderr"
    cmd2 = f"{cmd} > {stdout_fn} 2> {stderr_fn}"

    assert not non_blocking, "Getting output doesn't work with non_blocking"
    status = self.run(cmd2, False, ignore_errors=True)
    stdout = self.read(stdout_fn)
    stderr = self.read(stderr_fn)

    if self.last_status > 0:
      self.log(f"Warning: command '{cmd}' returned {status},"
               f" stdout was '{stdout}' stderr was '{stderr}'")
      if not ignore_errors:
        raise RuntimeError(f"Warning: command '{cmd}' returned {status},"
                           f" stdout was '{stdout}' stderr was '{stderr}'")

    return stdout, stderr

  def wait_for_file(self, fn: str, max_wait_sec: int = 3600 * 24 * 365,
                    check_interval: float = 0.02) -> bool:
    """
    Waits for file maximum of max_wait_sec. Returns True if file was detected within specified max_wait_sec
    Args:
      fn: filename on task machine
      max_wait_sec: how long to wait in seconds
      check_interval: how often to check in seconds
    Returns:
      False if waiting was was cut short by max_wait_sec limit, True otherwise
    """
    #    print("Waiting for file", fn)
    start_time = time.time()
    while True:
      if time.time() - start_time > max_wait_sec:
        util.log(f"Timeout exceeded ({max_wait_sec} sec) for {fn}")
        return False
      if not self.exists(fn):
        time.sleep(check_interval)
        continue
      else:
        break
    return True

  def _run_raw(self, cmd):
    """Runs command directly on every task in the job, skipping tmux interface. Use if want to create/manage additional tmux sessions manually."""
    raise NotImplementedError()

  def upload(self, local_fn: str, remote_fn: str = '',
             dont_overwrite: bool = False):
    """Uploads given file to the task. If remote_fn is not specified, dumps it
    into task current directory with the same name.

    Args:
      local_fn: location of file locally
      remote_fn: location of file on task
      dont_overwrite: if True, will be no-op if target file exists
      """
    raise NotImplementedError()

  def download(self, remote_fn: str, local_fn: str = ''):
    """Downloads remote file to current directory."""
    raise NotImplementedError()

  def write(self, fn, contents):
    """Write string contents to file fn in task."""
    raise NotImplementedError()

  def read(self, fn):
    """Read contents of file and return it as string."""
    raise NotImplementedError()

  def exists(self, fn) -> bool:
    """Checks if fn exists on task

    Args:
      fn: filename local to task
    Returns:
      true if fn exists on task machine
    """
    raise NotImplementedError()

  def log(self, message, *args):
    """Log to launcher console."""
    if args:
      message %= args

    print(f"{util.current_timestamp()} {self.name}: {message}")


class Job:
  name: str
  tasks: List[Task]

  #  run_: Run

  def __init__(self, name: str, tasks: List[Task] = None, **kwargs):
    """Initializes Job object, links tasks to refer back to the Job."""
    if tasks is None:
      tasks = []
    self.name = name
    self.tasks = tasks
    self.kwargs = kwargs
    # TODO: maybe backlinking is not needed
    for task in tasks:
      task.job = self

  @property
  def logdir(self):
    return self.tasks[0].logdir

  def _task_parallel(self, method, *args, **kwargs):
    """Runs given method on every task in the job in parallel. Blocks until all tasks finish. Propagates exception from first
    failed task."""

    exceptions = []

    def task_run(task):
      try:
        getattr(task, method)(*args, **kwargs)
      except Exception as e:
        exceptions.append(e)

    threads = [threading.Thread(name=f'task_{method}_{i}',
                                target=task_run, args=[t])
               for i, t in enumerate(self.tasks)]
    for thread in threads:
      thread.start()
    for thread in threads:
      thread.join()
    if exceptions:
      raise exceptions[0]

  def run(self, *args, **kwargs):
    """Runs command on every task in the job in parallel, blocks until all tasks finish.
    See Task for documentation of args/kwargs."""
    return self._task_parallel("run", *args, **kwargs)

  def propagate_env(self, *args, **kwargs):
    """See py:func:`aws_backend.Task.propagate_env`"""
    return self._task_parallel("propagate_env", *args, **kwargs)

  def run_with_output(self, *args, **kwargs):
    """Runs command on every task in the job in parallel, blocks until all tasks finish.
    See Task for documentation of args/kwargs."""
    return self._task_parallel("run_with_output", *args, **kwargs)

  def rsync(self, *args, **kwargs):
    """See :py:func:`backend.Task.rsync`"""
    return self._task_parallel("rsync", *args, **kwargs)

  def upload(self, *args, **kwargs):
    """See :py:func:`backend.Task.upload`"""
    return self._task_parallel("upload", *args, **kwargs)

  def write(self, *args, **kwargs):
    return self._task_parallel("write", *args, **kwargs)

  def _run_raw(self, *args, **kwargs):
    return self._task_parallel("_run_raw", *args, **kwargs)


# Implementation needs to be backend specific so that run.create_job calls backend-specific method
class Run:
  """Run is a collection of jobs that share state. IE, training run will contain gradient worker job, parameter
  server job, and TensorBoard visualizer job. These jobs will use the same shared directory to store checkpoints and
  event files.
  :ivar aws_placement_group_name: somedoc
  """
  jobs: List[Job]

  @property
  def logdir(self):
    raise NotImplementedError()

  # TODO: currently this is synchronous, use non_blocking wrapper like in Job to parallelize methods
  def run(self, *args, **kwargs):
    raise NotImplementedError()

  def run_with_output(self, *args, **kwargs):
    raise NotImplementedError()

  def _run_raw(self, *args, **kwargs):
    raise NotImplementedError()

  def upload(self, *args, **kwargs):
    raise NotImplementedError()

  def make_job(self, name='', **kwargs):
    raise NotImplementedError()


def make_task(**_kwargs):
  raise NotImplementedError()


def make_job(**_kwargs):
  raise NotImplementedError()


def make_run(**_kwargs):
  raise NotImplementedError()



#!/usr/bin/env python
#
# Deletes resources

import sys
import os
import argparse

from ncluster import aws_util as u
from ncluster import util

parser = argparse.ArgumentParser()
parser.add_argument('--kind', type=str, default='all',
                    help="which resources to delete, all/network/keypair/efs")
parser.add_argument('--force_delete_efs', action='store_true',
                    help="force deleting main EFS")
args = parser.parse_args()

EFS_NAME = u.get_prefix()
VPC_NAME = u.get_prefix()
SECURITY_GROUP_NAME = u.get_prefix()
ROUTE_TABLE_NAME = u.get_prefix()
KEYPAIR_NAME = u.get_keypair_name()

client = u.get_ec2_client()
ec2 = u.get_ec2_resource()


def response_type(response):
  return 'ok' if u.is_good_response(response) else 'failed'


def delete_efs():
  efss = u.get_efs_dict()
  efs_id = efss.get(EFS_NAME, '')
  efs_client = u.get_efs_client()
  if efs_id:
    try:
      # delete mount targets first
      print("About to delete %s (%s)" % (efs_id, EFS_NAME))
      response = efs_client.describe_mount_targets(FileSystemId=efs_id)
      assert u.is_good_response(response)
      for mount_response in response['MountTargets']:
        id_ = mount_response['MountTargetId']
        sys.stdout.write('Deleting mount target %s ... ' % (id_,))
        sys.stdout.flush()
        response = efs_client.delete_mount_target(MountTargetId=id_)
        print(response_type(response))

      sys.stdout.write('Deleting EFS %s (%s)... ' % (efs_id, EFS_NAME))
      sys.stdout.flush()
      u.delete_efs_by_id(efs_id)

    except Exception as e:
      sys.stdout.write(f'failed with {e}\n')
      util.log_error(str(e) + '\n')


def delete_network():
  if u.get_region() == 'us-east-1':
    util.log("(Internal safety switch. Not deleting resources in us-east-1, remove this line if you are really sure")
    return

  def delete_vpc(vpc, partial=True):
    """Deletes VPC + all resources, if "partial" set to True, only deletes associated security groups
    """
    print("Deleting VPC %s (%s) subresources:" % (VPC_NAME, vpc.id))

    # don't modify default VPC
    if not partial:
      for subnet in vpc.subnets.all():
        try:
          sys.stdout.write("Deleting subnet %s ... " % subnet.id)
          sys.stdout.write(response_type(subnet.delete()) + '\n')
        except Exception as e:
          sys.stdout.write('failed\n')
          util.log_error(str(e) + '\n')

      for gateway in vpc.internet_gateways.all():
        sys.stdout.write("Deleting gateway %s ... " % gateway.id)
        # note: if instances are using VPC, this fails with
        # botocore.exceptions.ClientError: An error occurred (DependencyViolation) when calling the DetachInternetGateway operation: Network vpc-ca4abab3 has some mapped public address(es). Please unmap those public address(es) before detaching the gateway.

        sys.stdout.write('detached ... ' if u.is_good_response(
          gateway.detach_from_vpc(VpcId=vpc.id)) else ' detach_failed ')
        sys.stdout.write('deleted ' if u.is_good_response(
          gateway.delete()) else ' delete_failed ')
        sys.stdout.write('\n')

      def desc():
        return "%s (%s)" % (route_table.id, u.get_name(route_table.tags))

      for route_table in vpc.route_tables.all():
        sys.stdout.write(f"Deleting route table {desc()} ... ")
        try:
          sys.stdout.write(response_type(route_table.delete()) + '\n')
        except Exception as e:
          sys.stdout.write('failed\n')
          util.log_error(str(e) + '\n')

    else:
      util.log(f"vpc {vpc.id} is a default VPC, only doing partial deletion")

    def desc():
      return "%s (%s, %s)" % (
        security_group.id, u.get_name(security_group.tags),
        security_group.group_name)

    ncluster_security_groups = u.get_security_group_names()
    for security_group in vpc.security_groups.all():
      # default group is undeletable, skip
      if security_group.group_name == 'default':
        continue

      # don't delete groups created outside of ncluster framework
      if security_group.group_name not in ncluster_security_groups:
        continue

      sys.stdout.write(
        'Deleting security group %s ... ' % (desc()))
      try:
        sys.stdout.write(response_type(security_group.delete()) + '\n')
      except Exception as e:
        sys.stdout.write('failed\n')
        util.log_error(str(e) + '\n')

    if not partial:
      sys.stdout.write("Deleting VPC %s ... " % vpc.id)
      try:
        sys.stdout.write(response_type(vpc.delete()) + '\n')
      except Exception as e:
        sys.stdout.write('failed\n')
        util.log_error(str(e) + '\n')

  existing_vpcs = u.get_vpc_dict()
  if VPC_NAME in existing_vpcs:
    # delete ncluster VPC
    delete_vpc(ec2.Vpc(existing_vpcs[VPC_NAME].id), partial=False)

  # delete ncluster resources on default VPC (partial=True)
  delete_vpc(u.get_default_vpc())


def delete_keypair():
  keypairs = u.get_keypair_dict()
  keypair = keypairs.get(KEYPAIR_NAME, '')
  if keypair:
    try:
      sys.stdout.write("Deleting keypair %s (%s) ... " % (keypair.key_name,
                                                          KEYPAIR_NAME))
      sys.stdout.write(response_type(keypair.delete()) + '\n')
    except Exception as e:
      sys.stdout.write('failed\n')
      util.log_error(str(e) + '\n')

  keypair_fn = u.get_keypair_fn()
  if os.path.exists(keypair_fn):
    print("Deleting local keypair file %s" % (keypair_fn,))
    os.system('rm -f ' + keypair_fn)


def delete_resources(force_delete_efs=False):
  region = os.environ['AWS_DEFAULT_REGION']

  resource = u.get_prefix()
  answer = input(f"Deleting resources for account {u.get_account_number()}:{u.get_account_name()}, region {u.get_region()}, sure? (y/N) ")

  if util.is_set("NCLUSTER_SKIP_CONFIRMATION"):
    print("NCLUSTER_SKIP_CONFIRMATION is set, skipping confirmation")
    answer = 'y'

  if not answer.lower() == "y":
    print("Didn't get y, doing nothing")
    return

  print(f"Deleting {resource} resources in region {region}")
  print(f"Make sure {resource} instances are terminated or this will fail.")

  if 'efs' in args.kind or 'all' in args.kind:
    if EFS_NAME == u.DEFAULT_PREFIX and not force_delete_efs:
      # this is default EFS, likely has stuff, require extra flag to delete it
      print("default EFS has useful stuff in it, not deleting it. Use force_delete_efs "
            "flag to force. This means security group deletion will fail as well.")
    else:
      delete_efs()
  if 'network' in args.kind or 'all' in args.kind:
    delete_network()
  if 'keypair' in args.kind or 'all' in args.kind:
    delete_keypair()


if __name__ == '__main__':
  delete_resources(force_delete_efs=args.force_delete_efs)
